#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# TODO: Missing doc strings
# pylint: disable=missing-module-docstring

__author__ = "John Eslick"

import pyomo.environ as pyo
from pyomo.network import Arc
from pyomo.common.config import ConfigBlock, ConfigValue, In, Bool

import idaes.core.util.constants as iconst
from idaes.core import UnitModelBlockData, declare_process_block_class
from idaes.core.util.config import is_physical_parameter_block
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.models.properties.modular_properties.base.generic_reaction import (
    GenericReactionParameterBlock,
)
import idaes.models.unit_models as um  # um = unit models
from idaes.core.util.initialization import propagate_state
import idaes.core.util.scaling as iscale
from idaes.core.solvers import get_solver
from idaes.models_extra.power_generation.properties.natural_gas_PR import (
    get_prop,
    get_rxn,
    EosType,
)
from idaes.core.util.exceptions import ConfigurationError, InitializationError
import idaes.logger as idaeslog
from idaes.core.util import from_json, to_json, StoreSpec


@declare_process_block_class("SoecDesign", doc="Simple SOEC model for process design.")
class SoecDesignData(UnitModelBlockData):
    """Simple 0D SOEC model. This model can be used to develop a flowsheet at the
    design point or for the basis of a surrogate. For off-design applications,
    a more detailed model is required."""

    CONFIG = UnitModelBlockData.CONFIG(implicit=True)
    CONFIG.declare(
        "oxygen_side_property_package",
        ConfigValue(
            domain=is_physical_parameter_block,
            description="Property package for the oxygen side.",
            doc=(
                "Property package for the oxygen side, using "
                "idaes.models_extra.power_generation.properties.natural_gas_PR is "
                "strongly recommended, either Peng-Robinson or Ideal is okay"
            ),
        ),
    )
    CONFIG.declare(
        "hydrogen_side_property_package",
        ConfigValue(
            domain=is_physical_parameter_block,
            description="Property package for the hydrogen side.",
            doc=(
                "Property package for the hydrogen side, using "
                "idaes.models_extra.power_generation.properties.natural_gas_PR is "
                "strongly recommended, either Peng-Robinson or Ideal is okay"
            ),
        ),
    )
    CONFIG.declare(
        "oxygen_side_property_package_args",
        ConfigBlock(
            implicit=True,
            description="Property package arguments for the oxygen side.",
            doc="Property package arguments for the oxygen side.",
        ),
    )
    CONFIG.declare(
        "hydrogen_side_property_package_args",
        ConfigBlock(
            implicit=True,
            description="Property package arguments for the hydrogen side.",
            doc="Property package arguments for the hydrogen side.",
        ),
    )
    CONFIG.declare(
        "reaction_eos",
        ConfigValue(
            default=EosType.PR,
            domain=In(EosType),
            description="Physical properties for electrolysis reactions",
            doc=(
                "Reaction properties equation of state in: "
                "{EosType.PR, EosType.IDEAL}."
            ),
        ),
    )
    CONFIG.declare(
        "has_heat_transfer",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Heat transfer term construction flag",
            doc="Indicates whether the SOEC is adiabatic. Default=False (adiabatic)",
        ),
    )

    def build(self):
        """Construct a unit model data block"""
        super().build()

        # Do a little validation.
        # The hydrogen side needs water and hydrogen
        if set(self.config.hydrogen_side_property_package.component_list) != {
            "H2",
            "H2O",
        }:
            raise ConfigurationError(
                "SOEC hydrogen side must contain exactly H2 and H2O"
            )
        # The oxygen side must have oxygen in it.
        if not "O2" in self.config.oxygen_side_property_package.component_list:
            raise ConfigurationError("SOEC oxygen side must contain O2")

        # build the block, use smaller methods to make it easier to follow
        self._add_electrolysis_properties()
        self._add_unit_models()
        self._add_arcs()
        self._add_variables()
        self._add_constraints()
        self._add_o2_translator_constraints()
        self._add_h2_inlet_translator_constraints()
        self._add_h2_outlet_translator_constraints()
        self._add_ports()
        self._scaling_guess()

    def _add_electrolysis_properties(self):
        """Add the electrolysis property package.  This has H2, H2O and O2,
        and is only used internally by the unit for the electrolysis reaction.
        """
        self.electrolysis_prop_params = GenericParameterBlock(
            **get_prop(
                {"H2O", "H2", "O2"}, phases={"Vap"}, eos=self.config.reaction_eos
            ),
            doc="Physical property parameters for the electrolysis reaction",
        )
        # Electrolysis is just the hydrogen combustion reaction backward
        self.electrolysis_rxn_params = GenericReactionParameterBlock(
            **get_rxn(self.electrolysis_prop_params, {"h2_cmb"}),
            doc="Reaction parameters",
        )

    def _add_unit_models(self):
        """Add the unit models that make up the composite model. The main
        things happening are 1) Electrolysis Reaction, 2) O2 Separation, 3) O2
        mixes with sweep gas, and 4) sweep gas heat transfer.  There are
        translator blocks to switch between property package depending on the
        existence of components in various streams.
        """
        self.h2_inlet_translator = um.Translator(
            doc="Translate hydrogen side properties to electrolysis properties",
            inlet_property_package=self.config.hydrogen_side_property_package,
            inlet_property_package_args=self.config.hydrogen_side_property_package_args,
            outlet_property_package=self.electrolysis_prop_params,
            outlet_state_defined=False,
        )
        self.electrolysis_reactor = um.StoichiometricReactor(
            doc="Electrolysis reactor",
            property_package=self.electrolysis_prop_params,
            reaction_package=self.electrolysis_rxn_params,
            has_pressure_change=False,
            has_heat_transfer=True,
        )
        self.o2_separator = um.Separator(
            property_package=self.electrolysis_prop_params,
            split_basis=um.SplittingType.componentFlow,
            outlet_list=["h2_strm", "o2_strm"],
        )
        self.h2_outlet_translator = um.Translator(
            doc="Translate electrolysis properties to hydrogen properties",
            inlet_property_package=self.electrolysis_prop_params,
            outlet_property_package=self.config.hydrogen_side_property_package,
            outlet_property_package_args=self.config.hydrogen_side_property_package_args,
            outlet_state_defined=False,
        )
        self.o2_translator = um.Translator(
            doc="Translate electrolysis properties to oxygen properties",
            inlet_property_package=self.electrolysis_prop_params,
            outlet_property_package=self.config.oxygen_side_property_package,
            outlet_property_package_args=self.config.oxygen_side_property_package_args,
            outlet_state_defined=False,
        )
        self.o2_mixer = um.Mixer(
            property_package=self.config.oxygen_side_property_package,
            property_package_args=self.config.oxygen_side_property_package_args,
            momentum_mixing_type=um.MomentumMixingType.none,
            inlet_list=["sweep_strm", "o2_strm"],
        )
        self.sweep_heater = um.Heater(
            property_package=self.config.oxygen_side_property_package,
            property_package_args=self.config.oxygen_side_property_package_args,
        )

    def _add_arcs(self):
        """Add streams to connect internal units."""
        self.strm1 = Arc(
            doc="Hydrogen side inlet to electrolysis translator",
            source=self.h2_inlet_translator.outlet,
            destination=self.electrolysis_reactor.inlet,
        )
        self.strm2 = Arc(
            doc="Electrolysis reactor to oxygen separation",
            source=self.electrolysis_reactor.outlet,
            destination=self.o2_separator.inlet,
        )
        self.strm3 = Arc(
            doc="Oxygen separation to hydrogen side outlet translator",
            source=self.o2_separator.h2_strm,
            destination=self.h2_outlet_translator.inlet,
        )
        self.strm4 = Arc(
            doc="Oxygen from electrolysis to translator to sweep properties",
            source=self.o2_separator.o2_strm,
            destination=self.o2_translator.inlet,
        )
        self.strm5 = Arc(
            doc="Oxygen translator to O2/sweep mixer",
            source=self.o2_translator.outlet,
            destination=self.o2_mixer.o2_strm,
        )
        self.strm6 = Arc(
            doc="O2/sweep mixer to sweep heater",
            source=self.o2_mixer.outlet,
            destination=self.sweep_heater.inlet,
        )
        # Probably don't need this and may need to call it again for dynamic
        # models, but this will make it easy for a user to have a flowsheet
        # with one steady state model in it.
        pyo.TransformationFactory("network.expand_arcs").apply_to(self)

    def _add_variables(self):
        """Add variables for unit model level constraints"""
        self.current = pyo.Var(
            self.flowsheet().time,
            initialize=0,
            units=pyo.units.ampere,
            doc="SOEC module electric current",
        )
        self.water_utilization = pyo.Var(
            self.flowsheet().time,
            units=pyo.units.dimensionless,
            doc="Fraction of water used in electrolysis reaction",
        )
        self.cell_potential = pyo.Var(
            self.flowsheet().time,
            initialize=1.3,
            units=pyo.units.volts,
            doc="Electric potential for a single cell",
        )
        self.heat = pyo.Var(
            self.flowsheet().time,
            units=pyo.units.W,
            initialize=0,
            doc="Heat transferred to the module",
        )
        self.hydrogen_side_outlet_temperature = pyo.Reference(
            self.electrolysis_reactor.control_volume.properties_out[:].temperature
        )
        self.oxygen_side_outlet_temperature = pyo.Reference(
            self.sweep_heater.control_volume.properties_out[:].temperature
        )

    def _add_constraints(self):
        """Add unit model constraints"""

        @self.Constraint(self.flowsheet().time)
        def water_utilization_eqn(b, t):
            return b.electrolysis_reactor.control_volume.properties_out[
                t
            ].flow_mol_comp[
                "H2O"
            ] == b.electrolysis_reactor.control_volume.properties_in[
                t
            ].flow_mol_comp[
                "H2O"
            ] * (
                1.0 - b.water_utilization[t]
            )

        @self.Expression(self.flowsheet().time)
        def current_expr(b, t):
            return (
                (
                    (
                        b.electrolysis_reactor.control_volume.properties_in[
                            t
                        ].mole_frac_comp["H2O"]
                        * b.electrolysis_reactor.control_volume.properties_in[
                            t
                        ].flow_mol
                    )
                    - (
                        b.electrolysis_reactor.control_volume.properties_out[
                            t
                        ].mole_frac_comp["H2O"]
                        * b.electrolysis_reactor.control_volume.properties_out[
                            t
                        ].flow_mol
                    )
                )
                * 2
                * iconst.Constants.faraday_constant
            )

        @self.Constraint(self.flowsheet().time)
        def current_eqn(b, t):
            return b.current[t] == b.current_expr[t]

        # Fix the O2 separator split fractions to just remove O2 from fuel side
        self.o2_separator.split_fraction[:, "o2_strm", "O2"].fix(1)
        self.o2_separator.split_fraction[:, "o2_strm", "H2O"].fix(0)
        self.o2_separator.split_fraction[:, "o2_strm", "H2"].fix(0)

        @self.o2_mixer.Constraint(self.flowsheet().time)
        def pressure_eqn(b, t):
            return b.mixed_state[t].pressure == b.sweep_strm_state[t].pressure

        if not self.config.has_heat_transfer:

            @self.Constraint(self.flowsheet().time)
            def heat_transfer_eqn(b, t):
                return b.heat[t] == 0

        @self.Expression(self.flowsheet().time)
        def delta_enth(b, t):
            return (
                b.o2_separator.h2_strm_state[t].get_enthalpy_flow_terms("Vap")
                + b.sweep_heater.control_volume.properties_out[
                    t
                ].get_enthalpy_flow_terms("Vap")
                - b.electrolysis_reactor.control_volume.properties_in[
                    t
                ].get_enthalpy_flow_terms("Vap")
                - b.o2_mixer.sweep_strm_state[t].get_enthalpy_flow_terms("Vap")
                + b.heat[t]
            )

        @self.Constraint(self.flowsheet().time)
        def cell_potential_eqn(b, t):
            return b.cell_potential[t] == b.delta_enth[t] / b.current[t]

    @staticmethod
    def _translator_ftp_constraints(translator):
        """Generalize some translator block constraints"""

        @translator.Constraint(translator.flowsheet().time)
        def temperature_eqn(b, t):
            return b.properties_out[t].temperature == b.properties_in[t].temperature

        @translator.Constraint(translator.flowsheet().time)
        def flow_mol_eqn(b, t):
            return b.properties_out[t].flow_mol == b.properties_in[t].flow_mol

        @translator.Constraint(translator.flowsheet().time)
        def pressure_eqn(b, t):
            return b.properties_out[t].pressure == b.properties_in[t].pressure

    def _add_o2_translator_constraints(self):
        """Add constraints to translate from electrolysis properties to sweep
        properties.  The stream being translated contains only O2.
        """
        t0 = self.flowsheet().time.first()
        comps = set(self.o2_translator.properties_out[t0].mole_frac_comp.keys())
        comps.remove("O2")

        @self.o2_translator.Constraint(self.flowsheet().time, comps)
        def mole_frac_comp_eqn(b, t, c):
            return b.properties_out[t].mole_frac_comp[c] == 1e-19

        self.o2_translator.properties_out[:].mole_frac_comp["O2"] = 1

        self._translator_ftp_constraints(self.o2_translator)

    def _add_h2_inlet_translator_constraints(self):
        """Translate the inlet hydrogen stream to electrolysis properties. This
        stream contains only H2 and H2O.
        """
        t0 = self.flowsheet().time.first()
        # There is no O2.  So set the value to zero.  Don't need a constraint
        # for this due to the sum = 1 constraint
        self.o2_translator.properties_out[:].mole_frac_comp["O2"] = 1e-19

        # Add constraints for the other components
        comps = set(self.h2_inlet_translator.properties_in[t0].mole_frac_comp.keys())

        @self.h2_inlet_translator.Constraint(self.flowsheet().time, comps)
        def mole_frac_comp_eqn(b, t, c):
            return (
                b.properties_out[t].mole_frac_comp[c]
                == b.properties_in[t].mole_frac_comp[c]
            )

        self._translator_ftp_constraints(self.h2_inlet_translator)

    def _add_h2_outlet_translator_constraints(self):
        """Translate the electrolysis properties to hydrogen side properties.
        This stream contains only H2 and H2O.
        """
        t0 = self.flowsheet().time.first()
        # Add constraints for the other components
        comps = set(self.h2_outlet_translator.properties_out[t0].mole_frac_comp.keys())
        comps.remove("H2")  # Need to remove one component due to sum = 1 constraint

        @self.h2_outlet_translator.Constraint(self.flowsheet().time, comps)
        def mole_frac_comp_eqn(b, t, c):
            return (
                b.properties_out[t].mole_frac_comp[c]
                == b.properties_in[t].mole_frac_comp[c]
            )

        self._translator_ftp_constraints(self.h2_outlet_translator)

    def _add_ports(self):
        """Add unit level ports"""
        self.add_inlet_port(
            name="hydrogen_side_inlet",
            block=self.h2_inlet_translator.properties_in,
            doc="Hydrogen side inlet port",
        )
        self.add_inlet_port(
            name="oxygen_side_inlet",
            block=self.o2_mixer.sweep_strm_state,
            doc="Oxygen side inlet port",
        )
        self.add_outlet_port(
            name="hydrogen_side_outlet",
            block=self.h2_outlet_translator.properties_out,
            doc="Hydrogen side outlet port",
        )
        self.add_outlet_port(
            name="oxygen_side_outlet",
            block=self.sweep_heater.control_volume.properties_out,
            doc="Oxygen side outlet port",
        )

    def set_flow_scale(self, scale=1):
        """Set default flow scaling in the property packages based on the
        expected magnitude of typical flows

        Args:
            scale (float): Expected molar flow variable scale

        Returns:
            None
        """
        self.config.oxygen_side_property_package.set_default_scaling("flow_mol", scale)
        self.config.hydrogen_side_property_package.set_default_scaling(
            "flow_mol", scale
        )
        self.electrolysis_prop_params.set_default_scaling("flow_mol", scale)

        self.config.oxygen_side_property_package.set_default_scaling(
            "flow_mol_phase", scale
        )
        self.config.hydrogen_side_property_package.set_default_scaling(
            "flow_mol_phase", scale
        )
        self.electrolysis_prop_params.set_default_scaling("flow_mol_phase", scale)

        for v in self.electrolysis_reactor.control_volume.rate_reaction_extent.values():
            iscale.set_scaling_factor(v, scale)
        for (
            v
        ) in self.electrolysis_reactor.control_volume.rate_reaction_generation.values():
            iscale.set_scaling_factor(v, scale)

    def set_heat_scale(self, scale=1e-5):
        """Set the heat transfer scale factor roughly based on the size of the
        process.

        Args:
            scale (float): Expected heat transfer scale

        Returns:
            None
        """
        iscale.set_scaling_factor(self.electrolysis_reactor.control_volume.heat, scale)
        iscale.set_scaling_factor(self.sweep_heater.control_volume.heat, scale)
        iscale.set_scaling_factor(self.heat, scale)

    def set_current_scale(self, scale=1e-6):
        """Set the expected electrical current scale

        Args:
            scale (float): Expected electrical current scale

        Returns:
            None
        """
        iscale.set_scaling_factor(self.current, scale)

    def _scaling_guess(self):
        self.set_flow_scale()
        self.set_heat_scale()
        self.set_current_scale()

        # For components that exist we don't expect particularly low
        # concentrations, so change the default to 10, which should be good
        # for mole fractions > 0.01 which is what we expect here.
        self.config.oxygen_side_property_package.set_default_scaling(
            "mole_frac_comp", 10
        )
        self.config.hydrogen_side_property_package.set_default_scaling(
            "mole_frac_comp", 10
        )
        self.electrolysis_prop_params.set_default_scaling("mole_frac_comp", 10)
        self.config.oxygen_side_property_package.set_default_scaling(
            "mole_frac_phase_comp", 10
        )
        self.config.hydrogen_side_property_package.set_default_scaling(
            "mole_frac_phase_comp", 10
        )
        self.electrolysis_prop_params.set_default_scaling("mole_frac_phase_comp", 10)

        # Set some other scale factors that we have a good guess for
        unt = self.electrolysis_reactor
        iscale.set_scaling_factor(
            unt.control_volume.properties_in[0.0].enth_mol_phase["Vap"], 1e-4
        )
        iscale.set_scaling_factor(
            unt.control_volume.properties_out[0.0].enth_mol_phase["Vap"], 1e-4
        )
        unt = self.o2_separator
        iscale.set_scaling_factor(unt.h2_strm_state[0.0].enth_mol_phase["Vap"], 1e-4)
        unt = self.o2_mixer
        iscale.set_scaling_factor(unt.o2_strm_state[0.0].enth_mol_phase["Vap"], 1e-4)
        iscale.set_scaling_factor(unt.o2_strm_state[0.0].enth_mol_phase["Vap"], 1e-4)
        iscale.set_scaling_factor(unt.mixed_state[0.0].enth_mol_phase["Vap"], 1e-4)
        iscale.set_scaling_factor(unt.sweep_strm_state[0.0].enth_mol_phase["Vap"], 1e-4)
        unt = self.sweep_heater
        iscale.set_scaling_factor(
            unt.control_volume.properties_in[0.0].enth_mol_phase["Vap"], 1e-4
        )
        iscale.set_scaling_factor(
            unt.control_volume.properties_out[0.0].enth_mol_phase["Vap"], 1e-4
        )

    def calculate_scaling_factors(self):
        """Calculate scale factors for the unit model equations"""
        for t in self.flowsheet().time:
            iscale.constraint_scaling_transform(
                self.water_utilization_eqn[t],
                iscale.get_scaling_factor(
                    self.electrolysis_reactor.control_volume.properties_in[t].flow_mol
                ),
            )
            iscale.constraint_scaling_transform(
                self.current_eqn[t], iscale.get_scaling_factor(self.current[t])
            )
            iscale.constraint_scaling_transform(
                self.o2_mixer.pressure_eqn[t],
                iscale.get_scaling_factor(self.o2_mixer.mixed_state[t].pressure),
            )
            iscale.constraint_scaling_transform(
                self.o2_translator.pressure_eqn[t],
                iscale.get_scaling_factor(self.o2_translator.properties_in[t].pressure),
            )
            iscale.constraint_scaling_transform(
                self.o2_translator.temperature_eqn[t],
                iscale.get_scaling_factor(
                    self.o2_translator.properties_in[t].temperature
                ),
            )
            iscale.constraint_scaling_transform(
                self.o2_translator.temperature_eqn[t],
                iscale.get_scaling_factor(
                    self.o2_translator.properties_in[t].temperature
                ),
            )
            iscale.constraint_scaling_transform(
                self.o2_translator.flow_mol_eqn[t],
                iscale.get_scaling_factor(self.o2_translator.properties_in[t].flow_mol),
            )
            iscale.constraint_scaling_transform(
                self.h2_inlet_translator.pressure_eqn[t],
                iscale.get_scaling_factor(
                    self.h2_inlet_translator.properties_in[t].pressure
                ),
            )
            iscale.constraint_scaling_transform(
                self.h2_inlet_translator.temperature_eqn[t],
                iscale.get_scaling_factor(
                    self.h2_inlet_translator.properties_in[t].temperature
                ),
            )
            iscale.constraint_scaling_transform(
                self.h2_inlet_translator.flow_mol_eqn[t],
                iscale.get_scaling_factor(
                    self.h2_inlet_translator.properties_in[t].flow_mol
                ),
            )
            iscale.constraint_scaling_transform(
                self.h2_outlet_translator.pressure_eqn[t],
                iscale.get_scaling_factor(
                    self.h2_outlet_translator.properties_in[t].pressure
                ),
            )
            iscale.constraint_scaling_transform(
                self.h2_outlet_translator.temperature_eqn[t],
                iscale.get_scaling_factor(
                    self.h2_outlet_translator.properties_in[t].temperature
                ),
            )
            iscale.constraint_scaling_transform(
                self.h2_outlet_translator.flow_mol_eqn[t],
                iscale.get_scaling_factor(
                    self.h2_outlet_translator.properties_in[t].flow_mol
                ),
            )
            iscale.constraint_scaling_transform(self.cell_potential_eqn[t], 1)
            if not self.config.has_heat_transfer:
                iscale.constraint_scaling_transform(
                    self.heat_transfer_eqn[t], iscale.get_scaling_factor(self.heat[t])
                )
        for (t, i), c in self.o2_translator.mole_frac_comp_eqn.items():
            iscale.constraint_scaling_transform(
                c,
                iscale.get_scaling_factor(
                    self.o2_translator.properties_out[t].mole_frac_comp[i]
                ),
            )
        for (t, i), c in self.h2_inlet_translator.mole_frac_comp_eqn.items():
            iscale.constraint_scaling_transform(
                c,
                iscale.get_scaling_factor(
                    self.h2_inlet_translator.properties_out[t].mole_frac_comp[i]
                ),
            )
        for (t, i), c in self.h2_outlet_translator.mole_frac_comp_eqn.items():
            iscale.constraint_scaling_transform(
                c,
                iscale.get_scaling_factor(
                    self.h2_outlet_translator.properties_out[t].mole_frac_comp[i]
                ),
            )

    def initialize_build(
        self,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        init_log.info_high("SOEC Initialization Starting")
        solver_obj = get_solver(solver, optarg)

        sp = StoreSpec.value_isfixed_isactive(only_fixed=True)
        istate = to_json(self, return_dict=True, wts=sp)

        self.oxygen_side_inlet.fix()
        self.hydrogen_side_inlet.fix()
        self.oxygen_side_outlet.unfix()
        self.hydrogen_side_outlet.unfix()

        self.h2_inlet_translator.initialize(outlvl=outlvl, solver=solver, optarg=optarg)
        propagate_state(self.strm1)
        self.electrolysis_reactor.initialize(
            outlvl=outlvl, solver=solver, optarg=optarg
        )
        propagate_state(self.strm2)
        self.o2_separator.initialize(outlvl=outlvl, solver=solver, optarg=optarg)
        propagate_state(self.strm3)
        self.h2_outlet_translator.initialize(
            outlvl=outlvl, solver=solver, optarg=optarg
        )
        propagate_state(self.strm4)
        self.o2_translator.initialize(outlvl=outlvl, solver=solver, optarg=optarg)
        propagate_state(self.strm5)
        self.o2_mixer.initialize(outlvl=outlvl, solver=solver, optarg=optarg)
        propagate_state(self.strm6)
        self.sweep_heater.initialize(outlvl=outlvl, solver=solver, optarg=optarg)

        for t in self.current.keys():
            self.current[t] = pyo.value(self.current_expr[t])

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solver_obj.solve(self, tee=slc.tee)
        if not pyo.check_optimal_termination(res):
            raise InitializationError("SOEC failed to initialize.")

        from_json(self, sd=istate, wts=sp)
        init_log.info_high("SOEC Initialization Complete")



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
A simple unit model to calculate power required to achieve isothermal
gas compression.

Data sources:
    [1] Stochastic Optimal Control Model for Natural Gas Network
        Operations. V. Zavala, 2014, Comp. Chem. Eng.
"""
# TODO: Missing docstrings
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

from pyomo.common.config import ConfigValue
from pyomo.core.base.constraint import Constraint
from pyomo.core.base.var import Var
from pyomo.core.base.units_container import units as pyunits

from idaes.core import (
    declare_process_block_class,
    MaterialBalanceType,
    UnitModelBlockData,
)
from idaes.core.util.config import is_physical_parameter_block


@declare_process_block_class("IsothermalCompressor")
class IsothermalCompressorData(UnitModelBlockData):
    CONFIG = UnitModelBlockData.CONFIG()

    CONFIG.declare(
        "property_package",
        ConfigValue(default=None, domain=is_physical_parameter_block),
    )

    def build(self):
        super(IsothermalCompressorData, self).build()

        time = self.flowsheet().time
        property_package = self.config.property_package

        if len(property_package.phase_list) != 1:
            raise ValueError(
                "%s can only be constructed with a "
                "single-phase property package.\n"
                "Got phases %s."
                % (self.__class__, [p for p in property_package.phase_list])
            )
        self.phase = next(iter(property_package.phase_list))
        if self.phase != "Vap":
            raise ValueError(
                '%s can only be constructed with a single phase, "Vap".'
                "Got phase %s." % (self.__class__, self.phase)
            )

        inlet_config = {"defined_state": True}
        self.inlet_state = property_package.build_state_block(time, **inlet_config)
        self.outlet_state = property_package.build_state_block(time)

        # A little annoying that add_port assumes the state block is indexed
        # by exactly time. If we supported unindexed state blocks here,
        # this entire unit model could be agnostic of time.
        # (This is not strictly true because add_state_material_balances
        # writes time-indexed balance equations, but we could get around this
        # by writing our own balance equations, which is easy enough.)
        self.add_port(
            name="inlet_port",
            block=self.inlet_state,
            doc="The inlet to the compressor",
        )
        self.add_port(
            name="outlet_port",
            block=self.outlet_state,
            doc="The outlet from the compressor",
        )

        self.add_state_material_balances(
            # TODO: Does componentPhase vs. componentTotal matter here?
            MaterialBalanceType.componentPhase,
            self.inlet_state,
            self.outlet_state,
        )

        self.add_state_isothermal_equation(self.inlet_state, self.outlet_state)

        # Add boost pressure and pressure equation
        self.boost_pressure = Var(
            time,
            initialize=0.0,
            units=pyunits.bar,
            # TODO: Should this variable not be initialized to its bound?
            bounds=(0.0, None),
            doc="Increase in pressure provided by this compressor",
        )
        self.add_pressure_change_equation(self.inlet_state, self.outlet_state)

        # Add compression coefficient
        self.beta = Var(
            time,
            initialize=0.28,  # Approximately (1.4 - 1.0)/1.4
            units=pyunits.dimensionless,
            bounds=(0.0, None),
            doc="Compression coefficient, beta.",
        )
        self.add_beta_equation(self.inlet_state)

        # Add power requirement
        self.power = Var(
            time,
            initialize=0.0,
            units=pyunits.kW,
            bounds=(0.0, None),
            doc="Work required to achieve the desired compression",
        )
        self.add_power_equation(self.inlet_state)

    def add_state_isothermal_equation(self, state1, state2):
        time = self.flowsheet().time

        # NOTE: We assume our states are only indexed by time and that
        # temperature is unindexed.
        def isothermal_rule(b, t):
            return state1[t].temperature == state2[t].temperature

        self.state_isothermal_eqn = Constraint(time, rule=isothermal_rule)

    def add_pressure_change_equation(self, inlet, outlet):
        time = self.flowsheet().time

        def pressure_rule(b, t):
            # Here we are using some quantities from the state blocks
            # and some from the unit model, so we must convert units.
            return pyunits.convert(inlet[t].pressure, pyunits.bar) + pyunits.convert(
                self.boost_pressure[t], pyunits.bar
            ) == pyunits.convert(outlet[t].pressure, pyunits.bar)

        self.pressure_change_eqn = Constraint(time, rule=pressure_rule)

    def add_beta_equation(self, inlet_state):
        time = self.flowsheet().time
        p = next(iter(self.config.property_package.phase_list))

        def beta_rule(b, t):
            # We know that we are constructing this property package with a
            # single phase, but we use heat_capacity_ratio_phase nonetheless
            # for compatibility with "generic" property packages.
            gamma = inlet_state[t].heat_capacity_ratio_phase[p]
            return b.beta[t] == (gamma - 1.0) / gamma

        self.beta_eqn = Constraint(time, rule=beta_rule)

    def add_power_equation(self, inlet_state):
        """
        This is Equation 2.10 in [1]
        """
        time = self.flowsheet().time

        def power_rule(b, t):
            cp_mass = inlet_state[t].cp_mol / inlet_state[t].mw
            power_expr = (
                inlet_state[t].temperature
                * cp_mass
                * inlet_state[t].flow_mass
                * (
                    (
                        (
                            pyunits.convert(inlet_state[t].pressure, pyunits.bar)
                            + pyunits.convert(self.boost_pressure[t], pyunits.bar)
                        )
                        / pyunits.convert(inlet_state[t].pressure, pyunits.bar)
                    )
                    ** b.beta[t]
                    - 1.0
                )
            )
            return b.power[t] == pyunits.convert(power_expr, pyunits.kW)

        self.power_eqn = Constraint(time, rule=power_rule)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for KerasSurrogate
"""
import pytest

pytest.importorskip("tensorflow.keras", reason="tensorflow.keras not available")
pytest.importorskip("omlt", reason="omlt not available")

import os.path
import pandas as pd
from pyomo.common.fileutils import this_file_dir
from pyomo.common.tempfiles import TempfileManager
from pyomo.environ import (
    ConcreteModel,
    Var,
    SolverFactory,
    assert_optimal_termination,
    value,
    Objective,
)
from idaes.core.surrogate.keras_surrogate import KerasSurrogate, load_keras_json_hd5
from idaes.core.surrogate.surrogate_block import SurrogateBlock
from idaes.core.surrogate.sampling.scaling import OffsetScaler


rtol = 1e-4
atol = 1e-4


def create_keras_model(name="T_data_1_10_10_2_sigmoid", return_keras_model_only=True):
    #    keras_folder_name = os.path.join(this_file_dir(), 'data', 'keras_models', name)
    #    keras_model = keras.models.load_model(keras_folder_name)
    keras_folder_name = os.path.join(this_file_dir(), "data", "keras_models")
    keras_model = load_keras_json_hd5(keras_folder_name, name)
    if return_keras_model_only:
        return keras_model

    if name.startswith("T_data"):
        inputs_scaler = OffsetScaler(
            expected_columns=["Temperature_K"],
            offset_series=pd.Series({"Temperature_K": 369.873737}),
            factor_series=pd.Series({"Temperature_K": 5.776343}),
        )
        outputs_scaler = OffsetScaler(
            expected_columns=["EnthMol", "VapFrac"],
            offset_series=pd.Series({"EnthMol": 59519.621464, "VapFrac": 0.556033}),
            factor_series=pd.Series({"EnthMol": 14805.365235, "VapFrac": 0.432934}),
        )
        input_labels = ["Temperature_K"]
        output_labels = ["EnthMol", "VapFrac"]
        input_bounds = {"Temperature_K": (360, 380)}
    elif name.startswith("PT_data"):
        inputs_scaler = OffsetScaler(
            expected_columns=["Temperature_K", "Pressure_Pa"],
            offset_series=pd.Series(
                {"Temperature_K": 369.983611, "Pressure_Pa": 111421.319811}
            ),
            factor_series=pd.Series(
                {"Temperature_K": 5.836047, "Pressure_Pa": 5917.954504}
            ),
        )
        outputs_scaler = OffsetScaler(
            expected_columns=["EnthMol", "VapFrac"],
            offset_series=pd.Series({"EnthMol": 54599.629980, "VapFrac": 0.403307}),
            factor_series=pd.Series({"EnthMol": 14654.226615, "VapFrac": 0.430181}),
        )
        input_labels = ["Temperature_K", "Pressure_Pa"]
        output_labels = ["EnthMol", "VapFrac"]
        input_bounds = {
            "Temperature_K": (360.0, 380.0),
            "Pressure_Pa": (101325.0, 1.2 * 101325.0),
        }

    keras_surrogate = KerasSurrogate(
        keras_model=keras_model,
        input_labels=input_labels,
        output_labels=output_labels,
        input_bounds=input_bounds,
        input_scaler=inputs_scaler,
        output_scaler=outputs_scaler,
    )

    return keras_surrogate


@pytest.mark.unit
def test_KerasSurrogate_construction_exceptions():
    keras_model = create_keras_model(name="T_data_1_10_10_2_sigmoid")
    inputs_scaler = OffsetScaler(
        expected_columns=["Temperature_K"],
        offset_series=pd.Series({"Temperature_K": 369.977879}),
        factor_series=pd.Series({"Temperature_K": 5.833396}),
    )
    outputs_scaler = OffsetScaler(
        expected_columns=["EnthMol", "VapFrac"],
        offset_series=pd.Series({"EnthMol": 54582.841444, "VapFrac": 0.402814}),
        factor_series=pd.Series({"EnthMol": 14642.206469, "VapFrac": 0.429829}),
    )

    # test exceptions for mismatched labels
    with pytest.raises(ValueError) as excinfo:
        keras_surrogate = KerasSurrogate(
            keras_model=keras_model,
            input_labels=["Temperature"],
            output_labels=["EnthMol", "VapFrac"],
            input_bounds={"Temperature": (360, 380)},
            input_scaler=inputs_scaler,
            output_scaler=outputs_scaler,
        )
    assert (
        str(excinfo.value)
        == "KerasSurrogate created with input_labels that do not match the expected columns in the input_scaler.\n"
        "input_labels=['Temperature']\n"
        "input_scaler.expected_columns()=['Temperature_K']"
    )

    with pytest.raises(ValueError) as excinfo:
        keras_surrogate = KerasSurrogate(
            keras_model=keras_model,
            input_labels=["Temperature_K"],
            output_labels=["EnthMol", "VapFraction"],
            input_bounds={"Temperature_K": (360, 380)},
            input_scaler=inputs_scaler,
            output_scaler=outputs_scaler,
        )
    assert (
        str(excinfo.value)
        == "KerasSurrogate created with output_labels that do not match the expected columns in the output_scaler.\n"
        "output_labels=['EnthMol', 'VapFraction']\n"
        "output_scaler.expected_columns()=['EnthMol', 'VapFrac']"
    )

    with pytest.raises(ValueError) as excinfo:
        keras_surrogate = KerasSurrogate(
            keras_model=keras_model,
            input_labels=["Temperature_K"],
            output_labels=["EnthMol", "VapFraction"],
            input_bounds={"Temperature": (360, 380)},
            input_scaler=inputs_scaler,
            output_scaler=outputs_scaler,
        )
    assert (
        str(excinfo.value)
        == "The input_labels did not match the keys in input_bounds.\n"
        "input_bounds.keys(): ['Temperature']\n"
        "input_labels: ['Temperature_K']"
    )


@pytest.mark.unit
def test_keras_evaluate():
    x = pd.DataFrame({"Temperature_K": [365, 370, 375]})
    keras_surrogate = create_keras_model(
        name="T_data_1_10_10_2_sigmoid", return_keras_model_only=False
    )
    y = keras_surrogate.evaluate_surrogate(x)
    expected_y = pd.DataFrame(
        {
            "EnthMol": [43156.97282890834, 63948.72663184423, 74702.24552784403],
            "VapFrac": [0.07331117299318313, 0.6918295409849882, 1.000048027182579],
        }
    )
    pd.testing.assert_frame_equal(y, expected_y, rtol=rtol, atol=atol)

    keras_surrogate = create_keras_model(
        name="T_data_1_10_10_2_relu", return_keras_model_only=False
    )
    y = keras_surrogate.evaluate_surrogate(x)
    expected_y = pd.DataFrame(
        {
            "EnthMol": [40938.491053831465, 63519.26581801092, 74745.56414331967],
            "VapFrac": [0.005596130031824131, 0.6987006701708436, 0.9976383677420616],
        }
    )
    pd.testing.assert_frame_equal(y, expected_y, rtol=rtol, atol=atol)

    x = pd.DataFrame(
        {
            "Temperature_K": [360, 370, 380],
            "Pressure_Pa": [1.05 * 101325, 1.10 * 101325, 1.15 * 101325],
        }
    )
    keras_surrogate = create_keras_model(
        name="PT_data_2_10_10_2_sigmoid", return_keras_model_only=False
    )
    y = keras_surrogate.evaluate_surrogate(x)
    expected_y = pd.DataFrame(
        {
            "EnthMol": [40236.16508772767, 48622.339003030225, 75320.3854834533],
            "VapFrac": [-0.001982489345431271, 0.21764500665882233, 1.0025723664236068],
        }
    )
    pd.testing.assert_frame_equal(y, expected_y, rtol=rtol, atol=atol)

    keras_surrogate = create_keras_model(
        name="PT_data_2_10_10_2_relu", return_keras_model_only=False
    )
    y = keras_surrogate.evaluate_surrogate(x)
    expected_y = pd.DataFrame(
        {
            "EnthMol": [40138.906193322684, 48418.02972793587, 75247.95294182807],
            "VapFrac": [0.002497340527117309, 0.2117767154205442, 0.9989431696164608],
        }
    )
    pd.testing.assert_frame_equal(y, expected_y, rtol=rtol, atol=atol)


@pytest.mark.unit
@pytest.mark.skipif(not SolverFactory("ipopt").available(False), reason="no Ipopt")
def test_keras_surrogate_auto_creating_variables():
    ###
    # Test 1->2 sigmoid
    ###
    keras_surrogate = create_keras_model(
        name="T_data_1_10_10_2_sigmoid", return_keras_model_only=False
    )
    x_test = pd.DataFrame({"Temperature_K": [365]})
    y_test = keras_surrogate.evaluate_surrogate(x_test)

    # Test full-space
    m = ConcreteModel()
    m.obj = Objective(expr=1)
    m.surrogate = SurrogateBlock()
    m.surrogate.build_model(
        surrogate_object=keras_surrogate,
        formulation=KerasSurrogate.Formulation.FULL_SPACE,
    )
    m.surrogate.inputs["Temperature_K"].fix(365)
    solver = SolverFactory("ipopt")
    status = solver.solve(m, tee=True)
    assert_optimal_termination(status)

    y_test_pyomo = pd.DataFrame(
        {
            "EnthMol": [value(m.surrogate.outputs["EnthMol"])],
            "VapFrac": [value(m.surrogate.outputs["VapFrac"])],
        }
    )
    pd.testing.assert_frame_equal(y_test, y_test_pyomo, rtol=rtol, atol=atol)

    # Test reduced-space
    m = ConcreteModel()
    m.obj = Objective(expr=1)
    m.surrogate = SurrogateBlock()
    m.surrogate.build_model(
        surrogate_object=keras_surrogate,
        formulation=KerasSurrogate.Formulation.REDUCED_SPACE,
    )
    m.surrogate.inputs["Temperature_K"].fix(365)
    solver = SolverFactory("ipopt")
    status = solver.solve(m, tee=True)
    assert_optimal_termination(status)

    y_test_pyomo = pd.DataFrame(
        {
            "EnthMol": [value(m.surrogate.outputs["EnthMol"])],
            "VapFrac": [value(m.surrogate.outputs["VapFrac"])],
        }
    )
    pd.testing.assert_frame_equal(y_test, y_test_pyomo, rtol=rtol, atol=atol)

    ###
    # Test 1->2 relu
    ###
    keras_surrogate = create_keras_model(
        name="T_data_1_10_10_2_relu", return_keras_model_only=False
    )
    x_test = pd.DataFrame({"Temperature_K": [365]})
    y_test = keras_surrogate.evaluate_surrogate(x_test)

    # Test relu complementarity
    m = ConcreteModel()
    m.obj = Objective(expr=1)
    m.surrogate = SurrogateBlock()
    m.surrogate.build_model(
        surrogate_object=keras_surrogate,
        formulation=KerasSurrogate.Formulation.RELU_COMPLEMENTARITY,
    )
    m.surrogate.inputs["Temperature_K"].fix(365)
    solver = SolverFactory("ipopt")
    status = solver.solve(m, tee=True)
    assert_optimal_termination(status)

    y_test_pyomo = pd.DataFrame(
        {
            "EnthMol": [value(m.surrogate.outputs["EnthMol"])],
            "VapFrac": [value(m.surrogate.outputs["VapFrac"])],
        }
    )
    pd.testing.assert_frame_equal(y_test, y_test_pyomo, rtol=rtol, atol=atol)

    ###
    # Test 2->2 sigmoid
    ###
    keras_surrogate = create_keras_model(
        name="PT_data_2_10_10_2_sigmoid", return_keras_model_only=False
    )
    x_test = pd.DataFrame({"Temperature_K": [365], "Pressure_Pa": [1.1 * 101325]})
    y_test = keras_surrogate.evaluate_surrogate(x_test)

    # Test full-space
    m = ConcreteModel()
    m.obj = Objective(expr=1)
    m.surrogate = SurrogateBlock()
    m.surrogate.build_model(
        surrogate_object=keras_surrogate,
        formulation=KerasSurrogate.Formulation.FULL_SPACE,
    )
    m.surrogate.inputs["Temperature_K"].fix(365)
    m.surrogate.inputs["Pressure_Pa"].fix(1.1 * 101325)
    solver = SolverFactory("ipopt")
    status = solver.solve(m, tee=True)
    assert_optimal_termination(status)

    y_test_pyomo = pd.DataFrame(
        {
            "EnthMol": [value(m.surrogate.outputs["EnthMol"])],
            "VapFrac": [value(m.surrogate.outputs["VapFrac"])],
        }
    )
    pd.testing.assert_frame_equal(y_test, y_test_pyomo, rtol=rtol, atol=atol)

    # Test reduced-space
    m = ConcreteModel()
    m.obj = Objective(expr=1)
    m.surrogate = SurrogateBlock()
    m.surrogate.build_model(
        surrogate_object=keras_surrogate,
        formulation=KerasSurrogate.Formulation.REDUCED_SPACE,
    )
    m.surrogate.inputs["Temperature_K"].fix(365)
    m.surrogate.inputs["Pressure_Pa"].fix(1.1 * 101325)
    solver = SolverFactory("ipopt")
    status = solver.solve(m, tee=True)
    assert_optimal_termination(status)

    y_test_pyomo = pd.DataFrame(
        {
            "EnthMol": [value(m.surrogate.outputs["EnthMol"])],
            "VapFrac": [value(m.surrogate.outputs["VapFrac"])],
        }
    )
    pd.testing.assert_frame_equal(y_test, y_test_pyomo, rtol=rtol, atol=atol)

    ###
    # Test 2->2 relu
    ###
    keras_surrogate = create_keras_model(
        name="PT_data_2_10_10_2_relu", return_keras_model_only=False
    )
    x_test = pd.DataFrame({"Temperature_K": [365], "Pressure_Pa": [1.1 * 101325]})
    y_test = keras_surrogate.evaluate_surrogate(x_test)

    # Test relu complementarity
    m = ConcreteModel()
    m.obj = Objective(expr=1)
    m.surrogate = SurrogateBlock()
    m.surrogate.build_model(
        surrogate_object=keras_surrogate,
        formulation=KerasSurrogate.Formulation.RELU_COMPLEMENTARITY,
    )
    m.surrogate.inputs["Temperature_K"].fix(365)
    m.surrogate.inputs["Pressure_Pa"].fix(1.1 * 101325)
    solver = SolverFactory("ipopt")
    status = solver.solve(m, tee=True)
    assert_optimal_termination(status)

    y_test_pyomo = pd.DataFrame(
        {
            "EnthMol": [value(m.surrogate.outputs["EnthMol"])],
            "VapFrac": [value(m.surrogate.outputs["VapFrac"])],
        }
    )
    pd.testing.assert_frame_equal(y_test, y_test_pyomo, rtol=rtol, atol=atol)


@pytest.mark.unit
@pytest.mark.skipif(not SolverFactory("glpk").available(False), reason="no glpk")
def test_keras_surrogate_auto_creating_variables_glpk():
    ###
    # Test 1->2 relu
    ###
    keras_surrogate = create_keras_model(
        name="T_data_1_10_10_2_relu", return_keras_model_only=False
    )
    x_test = pd.DataFrame({"Temperature_K": [365]})
    y_test = keras_surrogate.evaluate_surrogate(x_test)

    # Test relu bigm
    m = ConcreteModel()
    m.obj = Objective(expr=1)
    m.surrogate = SurrogateBlock()
    m.surrogate.build_model(
        surrogate_object=keras_surrogate,
        formulation=KerasSurrogate.Formulation.RELU_BIGM,
    )
    m.surrogate.inputs["Temperature_K"].fix(365)
    solver = SolverFactory("glpk")
    status = solver.solve(m, tee=True)
    assert_optimal_termination(status)

    y_test_pyomo = pd.DataFrame(
        {
            "EnthMol": [value(m.surrogate.outputs["EnthMol"])],
            "VapFrac": [value(m.surrogate.outputs["VapFrac"])],
        }
    )
    pd.testing.assert_frame_equal(y_test, y_test_pyomo, rtol=rtol, atol=atol)

    ###
    # Test 2->2 relu
    ###
    keras_surrogate = create_keras_model(
        name="PT_data_2_10_10_2_relu", return_keras_model_only=False
    )
    x_test = pd.DataFrame({"Temperature_K": [365], "Pressure_Pa": [1.1 * 101325]})
    y_test = keras_surrogate.evaluate_surrogate(x_test)

    # Test relu bigm
    m = ConcreteModel()
    m.obj = Objective(expr=1)
    m.surrogate = SurrogateBlock()
    m.surrogate.build_model(
        surrogate_object=keras_surrogate,
        formulation=KerasSurrogate.Formulation.RELU_BIGM,
    )
    m.surrogate.inputs["Temperature_K"].fix(365)
    m.surrogate.inputs["Pressure_Pa"].fix(1.1 * 101325)
    solver = SolverFactory("glpk")
    status = solver.solve(m, tee=True)
    assert_optimal_termination(status)

    y_test_pyomo = pd.DataFrame(
        {
            "EnthMol": [value(m.surrogate.outputs["EnthMol"])],
            "VapFrac": [value(m.surrogate.outputs["VapFrac"])],
        }
    )
    pd.testing.assert_frame_equal(y_test, y_test_pyomo, rtol=rtol, atol=atol)


@pytest.mark.unit
@pytest.mark.skipif(not SolverFactory("ipopt").available(False), reason="no Ipopt")
def test_keras_surrogate_with_variables():
    keras_surrogate = create_keras_model(
        name="PT_data_2_10_10_2_sigmoid", return_keras_model_only=False
    )
    x_test = pd.DataFrame({"Temperature_K": [365], "Pressure_Pa": [1.1 * 101325]})
    y_test = keras_surrogate.evaluate_surrogate(x_test)

    # Test provide scalar inputs, auto create outputs
    m = ConcreteModel()
    m.obj = Objective(expr=1)
    m.T = Var()
    m.P = Var()
    m.surrogate = SurrogateBlock()
    m.surrogate.build_model(
        surrogate_object=keras_surrogate,
        input_vars=[m.T, m.P],
        formulation=KerasSurrogate.Formulation.FULL_SPACE,
    )
    m.T.fix(365)
    m.P.fix(1.1 * 101325)
    solver = SolverFactory("ipopt")
    status = solver.solve(m, tee=True)
    assert_optimal_termination(status)

    y_test_pyomo = pd.DataFrame(
        {
            "EnthMol": [value(m.surrogate.outputs["EnthMol"])],
            "VapFrac": [value(m.surrogate.outputs["VapFrac"])],
        }
    )
    pd.testing.assert_frame_equal(y_test, y_test_pyomo, rtol=rtol, atol=atol)

    # Test provide indexed inputs, auto create outputs
    m = ConcreteModel()
    m.obj = Objective(expr=1)
    m.inputs = Var(["T", "P"])
    m.surrogate = SurrogateBlock()
    m.surrogate.build_model(
        surrogate_object=keras_surrogate,
        input_vars=[m.inputs],
        formulation=KerasSurrogate.Formulation.FULL_SPACE,
    )
    m.inputs["T"].fix(365)
    m.inputs["P"].fix(1.1 * 101325)
    solver = SolverFactory("ipopt")
    status = solver.solve(m, tee=True)
    assert_optimal_termination(status)

    y_test_pyomo = pd.DataFrame(
        {
            "EnthMol": [value(m.surrogate.outputs["EnthMol"])],
            "VapFrac": [value(m.surrogate.outputs["VapFrac"])],
        }
    )
    pd.testing.assert_frame_equal(y_test, y_test_pyomo, rtol=rtol, atol=atol)

    # Test auto-create inputs, provide scalar outputs
    m = ConcreteModel()
    m.obj = Objective(expr=1)
    m.H = Var()
    m.V = Var()
    m.surrogate = SurrogateBlock()
    m.surrogate.build_model(
        surrogate_object=keras_surrogate,
        output_vars=[m.H, m.V],
        formulation=KerasSurrogate.Formulation.FULL_SPACE,
    )
    m.surrogate.inputs["Temperature_K"].fix(365)
    m.surrogate.inputs["Pressure_Pa"].fix(1.1 * 101325)
    solver = SolverFactory("ipopt")
    status = solver.solve(m, tee=True)
    assert_optimal_termination(status)

    y_test_pyomo = pd.DataFrame({"EnthMol": [value(m.H)], "VapFrac": [value(m.V)]})
    pd.testing.assert_frame_equal(y_test, y_test_pyomo, rtol=rtol, atol=atol)

    # Test auto-create inputs, provide indexed outputs
    m = ConcreteModel()
    m.obj = Objective(expr=1)
    m.outputs = Var(["H", "V"])
    m.surrogate = SurrogateBlock()
    m.surrogate.build_model(
        surrogate_object=keras_surrogate,
        output_vars=[m.outputs],
        formulation=KerasSurrogate.Formulation.FULL_SPACE,
    )
    m.surrogate.inputs["Temperature_K"].fix(365)
    m.surrogate.inputs["Pressure_Pa"].fix(1.1 * 101325)
    solver = SolverFactory("ipopt")
    status = solver.solve(m, tee=True)
    assert_optimal_termination(status)

    y_test_pyomo = pd.DataFrame(
        {"EnthMol": [value(m.outputs["H"])], "VapFrac": [value(m.outputs["V"])]}
    )
    pd.testing.assert_frame_equal(y_test, y_test_pyomo, rtol=rtol, atol=atol)

    # Test provide scalar inputs, provide indexed outputs
    m = ConcreteModel()
    m.obj = Objective(expr=1)
    m.T = Var()
    m.P = Var()
    m.outputs = Var(["H", "V"])
    m.surrogate = SurrogateBlock()
    m.surrogate.build_model(
        surrogate_object=keras_surrogate,
        input_vars=[m.T, m.P],
        output_vars=[m.outputs],
        formulation=KerasSurrogate.Formulation.FULL_SPACE,
    )
    m.T.fix(365)
    m.P.fix(1.1 * 101325)
    solver = SolverFactory("ipopt")
    status = solver.solve(m, tee=True)
    assert_optimal_termination(status)

    y_test_pyomo = pd.DataFrame(
        {"EnthMol": [value(m.outputs["H"])], "VapFrac": [value(m.outputs["V"])]}
    )
    pd.testing.assert_frame_equal(y_test, y_test_pyomo, rtol=rtol, atol=atol)


@pytest.mark.unit
@pytest.mark.skipif(not SolverFactory("ipopt").available(False), reason="no Ipopt")
def test_save_load():
    keras_surrogate = create_keras_model(
        name="PT_data_2_10_10_2_sigmoid", return_keras_model_only=False
    )

    new_keras_surrogate = None
    dname = None
    with TempfileManager.new_context() as tf:
        dname = tf.mkdtemp()
        keras_surrogate.save_to_folder(dname)
        assert os.path.isdir(dname)
        assert os.path.isfile(os.path.join(dname, "idaes_info.json"))

        new_keras_surrogate = KerasSurrogate.load_from_folder(dname)

    # Check for clean up
    assert not os.path.isdir(dname)

    # check surrogate data members
    assert new_keras_surrogate._input_labels == ["Temperature_K", "Pressure_Pa"]
    assert new_keras_surrogate._output_labels == ["EnthMol", "VapFrac"]
    assert sorted(new_keras_surrogate._input_bounds.keys()) == [
        "Pressure_Pa",
        "Temperature_K",
    ]
    assert new_keras_surrogate._input_bounds["Temperature_K"][0] == pytest.approx(360.0)
    assert new_keras_surrogate._input_bounds["Temperature_K"][1] == pytest.approx(380.0)
    assert new_keras_surrogate._input_bounds["Pressure_Pa"][0] == pytest.approx(
        101325.0
    )
    assert new_keras_surrogate._input_bounds["Pressure_Pa"][1] == pytest.approx(
        1.2 * 101325.0
    )

    # check input scaler
    expected_columns = ["Temperature_K", "Pressure_Pa"]
    offset_series = pd.Series(
        {"Temperature_K": 369.983611, "Pressure_Pa": 111421.319811}
    )
    factor_series = pd.Series({"Temperature_K": 5.836047, "Pressure_Pa": 5917.954504})
    scaler = new_keras_surrogate._input_scaler
    assert scaler._expected_columns == expected_columns
    pd.testing.assert_series_equal(scaler._offset, offset_series, rtol=rtol, atol=atol)
    pd.testing.assert_series_equal(scaler._factor, factor_series, rtol=rtol, atol=atol)

    # check output scaler
    expected_columns = ["EnthMol", "VapFrac"]
    offset_series = pd.Series({"EnthMol": 54599.629980, "VapFrac": 0.403307})
    factor_series = pd.Series({"EnthMol": 14654.226615, "VapFrac": 0.430181})
    scaler = new_keras_surrogate._output_scaler
    assert scaler._expected_columns == expected_columns
    pd.testing.assert_series_equal(scaler._offset, offset_series, rtol=rtol, atol=atol)
    pd.testing.assert_series_equal(scaler._factor, factor_series, rtol=rtol, atol=atol)

    # check evaluation
    x_test = pd.DataFrame(
        {
            "Temperature_K": [360, 370, 380],
            "Pressure_Pa": [1.05 * 101325, 1.10 * 101325, 1.15 * 101325],
        }
    )
    y_test = keras_surrogate.evaluate_surrogate(x_test)
    expected_y = pd.DataFrame(
        {
            "EnthMol": [40236.16508772767, 48622.339003030225, 75320.3854834533],
            "VapFrac": [-0.001982489345431271, 0.21764500665882233, 1.0025723664236068],
        }
    )
    pd.testing.assert_frame_equal(y_test, expected_y, rtol=rtol, atol=atol)

    # check solve with pyomo
    x_test = pd.DataFrame({"Temperature_K": [365], "Pressure_Pa": [1.1 * 101325]})
    y_test = keras_surrogate.evaluate_surrogate(x_test)

    m = ConcreteModel()
    m.obj = Objective(expr=1)
    m.surrogate = SurrogateBlock()
    m.surrogate.build_model(
        surrogate_object=keras_surrogate,
        formulation=KerasSurrogate.Formulation.FULL_SPACE,
    )
    m.surrogate.inputs["Temperature_K"].fix(365)
    m.surrogate.inputs["Pressure_Pa"].fix(1.1 * 101325)
    solver = SolverFactory("ipopt")
    status = solver.solve(m, tee=True)
    assert_optimal_termination(status)

    y_test_pyomo = pd.DataFrame(
        {
            "EnthMol": [value(m.surrogate.outputs["EnthMol"])],
            "VapFrac": [value(m.surrogate.outputs["VapFrac"])],
        }
    )
    pd.testing.assert_frame_equal(y_test, y_test_pyomo, rtol=rtol, atol=atol)


@pytest.mark.unit
@pytest.mark.skipif(not SolverFactory("ipopt").available(False), reason="no Ipopt")
def test_noscalers():
    keras_folder_name = os.path.join(this_file_dir(), "data", "keras_models")
    keras_model = load_keras_json_hd5(keras_folder_name, "PT_data_2_10_10_2_sigmoid")

    input_labels = ["Temperature_K", "Pressure_Pa"]
    output_labels = ["EnthMol", "VapFrac"]
    input_bounds = {"Temperature_K": (-3.0, 3.0), "Pressure_Pa": (-3.0, 3.0)}

    keras_surrogate = KerasSurrogate(
        keras_model=keras_model,
        input_labels=input_labels,
        output_labels=output_labels,
        input_bounds=input_bounds,
    )
    # check solve with pyomo
    x_test = pd.DataFrame({"Temperature_K": [0.5], "Pressure_Pa": [0.5]})
    y_test = keras_surrogate.evaluate_surrogate(x_test)

    m = ConcreteModel()
    m.obj = Objective(expr=1)
    m.surrogate = SurrogateBlock()
    m.surrogate.build_model(
        surrogate_object=keras_surrogate,
        formulation=KerasSurrogate.Formulation.FULL_SPACE,
    )
    m.surrogate.inputs["Temperature_K"].fix(0.5)
    m.surrogate.inputs["Pressure_Pa"].fix(0.5)
    solver = SolverFactory("ipopt")
    status = solver.solve(m, tee=True)
    assert_optimal_termination(status)

    y_test_pyomo = pd.DataFrame(
        {
            "EnthMol": [value(m.surrogate.outputs["EnthMol"])],
            "VapFrac": [value(m.surrogate.outputs["VapFrac"])],
        }
    )
    pd.testing.assert_frame_equal(
        y_test, y_test_pyomo, check_dtype=False, rtol=rtol, atol=atol
    )


@pytest.mark.unit
def test_invalid_formulation():
    keras_surrogate = create_keras_model(
        name="T_data_1_10_10_2_sigmoid", return_keras_model_only=False
    )
    m = ConcreteModel()
    m.obj = Objective(expr=1)
    m.surrogate = SurrogateBlock()
    with pytest.raises(ValueError) as excinfo:
        m.surrogate.build_model(surrogate_object=keras_surrogate, formulation="foo")
    assert (
        str(excinfo.value) == 'An unrecognized formulation "foo" was passed '
        "to KerasSurrogate.populate_block. Please pass a valid formulation."
    )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Modular methods for calculating bubble and dew points
"""
# TODO: Pylint complains about variables with _x names as they are built by other classes
# pylint: disable=protected-access

from pyomo.environ import Constraint

from idaes.models.properties.modular_properties.base.utility import (
    get_method,
    get_component_object as cobj,
    identify_VL_component_list,
)
import idaes.core.util.scaling as iscale
from idaes.core.util.exceptions import ConfigurationError


class IdealBubbleDew:
    """Bubble and dew point calculations for ideal systems."""

    # -------------------------------------------------------------------------
    # Bubble temperature methods
    # This approach can only be used when both liquid and vapor phases use
    # Ideal properties
    # Henry's Law components also cause issues due to the need to (potentially)
    # calculate concentrations at the bubble and dew points
    @staticmethod
    def temperature_bubble(b):
        """
        Rule for calculating bubble temperature
        """
        _non_vle_phase_check(b)
        try:

            def rule_bubble_temp(b, p1, p2):
                (
                    l_phase,
                    v_phase,
                    vl_comps,
                    henry_comps,
                    _,
                    v_only_comps,
                ) = identify_VL_component_list(b, (p1, p2))

                if l_phase is None or v_phase is None:
                    # Not a VLE pair
                    return Constraint.Skip
                elif v_only_comps != []:
                    # Non-condensables present, no bubble point
                    return Constraint.Skip

                return (
                    sum(
                        b.mole_frac_comp[j]
                        * get_method(b, "pressure_sat_comp", j)(
                            b, cobj(b, j), b.temperature_bubble[p1, p2]
                        )
                        for j in vl_comps
                    )
                    + sum(
                        b.mole_frac_comp[j]
                        * b.params.get_component(j)
                        .config.henry_component[l_phase]["method"]
                        .return_expression(b, l_phase, j, b.temperature_bubble[p1, p2])
                        for j in henry_comps
                    )
                    - b.pressure
                ) == 0

            b.eq_temperature_bubble = Constraint(
                b.params._pe_pairs, rule=rule_bubble_temp
            )
        except AttributeError:
            b.del_component(b.eq_temperature_bubble)
            raise

        # Don't need a try/except here, will pass if first constraint did
        def rule_mole_frac_bubble_temp(b, p1, p2, j):
            (
                l_phase,
                v_phase,
                vl_comps,
                henry_comps,
                _,
                v_only_comps,
            ) = identify_VL_component_list(b, (p1, p2))

            if l_phase is None or v_phase is None:
                # Not a VLE pair
                return Constraint.Skip
            elif v_only_comps != []:
                # Non-condensables present, no bubble point
                return Constraint.Skip

            if j in vl_comps:
                return b._mole_frac_tbub[p1, p2, j] * b.pressure == (
                    b.mole_frac_comp[j]
                    * get_method(b, "pressure_sat_comp", j)(
                        b, cobj(b, j), b.temperature_bubble[p1, p2]
                    )
                )
            elif j in henry_comps:
                return b._mole_frac_tbub[p1, p2, j] * b.pressure == (
                    b.mole_frac_comp[j]
                    * b.params.get_component(j)
                    .config.henry_component[l_phase]["method"]
                    .return_expression(b, l_phase, j, b.temperature_bubble[p1, p2])
                )
            else:
                return b._mole_frac_tbub[p1, p2, j] == 0

        b.eq_mole_frac_tbub = Constraint(
            b.params._pe_pairs, b.component_list, rule=rule_mole_frac_bubble_temp
        )

    @staticmethod
    def scale_temperature_bubble(b, overwrite=True):
        """
        Scaling method for bubble temperature
        """
        sf_P = iscale.get_scaling_factor(b.pressure, default=1e-5, warning=True)
        sf_mf = iscale.get_scaling_factor(b.mole_frac_comp, default=1e3, warning=True)

        for pp in b.params._pe_pairs:
            for j in b.component_list:
                (
                    l_phase,
                    v_phase,
                    vl_comps,
                    _,
                    _,
                    v_only_comps,
                ) = identify_VL_component_list(b, pp)
                if l_phase is None or v_phase is None:
                    continue
                elif v_only_comps != []:
                    continue

                if j in vl_comps:
                    sf = sf_P * sf_mf
                else:
                    sf = sf_mf

                iscale.constraint_scaling_transform(
                    b.eq_temperature_bubble[pp[0], pp[1]], sf_P, overwrite=overwrite
                )
                iscale.constraint_scaling_transform(
                    b.eq_mole_frac_tbub[pp[0], pp[1], j], sf, overwrite=overwrite
                )

    # -------------------------------------------------------------------------
    # Dew temperature methods
    @staticmethod
    def temperature_dew(b):
        """
        Method for calculating dew temperature
        """
        _non_vle_phase_check(b)
        try:

            def rule_dew_temp(b, p1, p2):
                (
                    l_phase,
                    v_phase,
                    vl_comps,
                    henry_comps,
                    l_only_comps,
                    _,
                ) = identify_VL_component_list(b, (p1, p2))

                if l_phase is None or v_phase is None:
                    # Not a VLE pair
                    return Constraint.Skip
                elif l_only_comps != []:
                    # Non-vaporisables present, no dew point
                    return Constraint.Skip

                return (
                    b.pressure
                    * (
                        sum(
                            b.mole_frac_comp[j]
                            / get_method(b, "pressure_sat_comp", j)(
                                b, cobj(b, j), b.temperature_dew[p1, p2]
                            )
                            for j in vl_comps
                        )
                        + sum(
                            b.mole_frac_comp[j]
                            / b.params.get_component(j)
                            .config.henry_component[l_phase]["method"]
                            .return_expression(b, l_phase, j, b.temperature_dew[p1, p2])
                            for j in henry_comps
                        )
                    )
                    - 1
                    == 0
                )

            b.eq_temperature_dew = Constraint(b.params._pe_pairs, rule=rule_dew_temp)
        except AttributeError:
            b.del_component(b.eq_temperature_dew)
            raise

        # Don't need a try/except here, will pass if first constraint did
        def rule_mole_frac_dew_temp(b, p1, p2, j):
            (
                l_phase,
                v_phase,
                vl_comps,
                henry_comps,
                l_only_comps,
                _,
            ) = identify_VL_component_list(b, (p1, p2))

            if l_phase is None or v_phase is None:
                # Not a VLE pair
                return Constraint.Skip
            elif l_only_comps != []:
                # Non-vaporisables present, no dew point
                return Constraint.Skip

            if j in vl_comps:
                return (
                    b._mole_frac_tdew[p1, p2, j]
                    * get_method(b, "pressure_sat_comp", j)(
                        b, cobj(b, j), b.temperature_dew[p1, p2]
                    )
                    == b.mole_frac_comp[j] * b.pressure
                )
            elif j in henry_comps:
                return (
                    b._mole_frac_tdew[p1, p2, j]
                    * b.params.get_component(j)
                    .config.henry_component[l_phase]["method"]
                    .return_expression(b, l_phase, j, b.temperature_dew[p1, p2])
                    == b.mole_frac_comp[j] * b.pressure
                )
            else:
                return b._mole_frac_tdew[p1, p2, j] == 0

        b.eq_mole_frac_tdew = Constraint(
            b.params._pe_pairs, b.component_list, rule=rule_mole_frac_dew_temp
        )

    @staticmethod
    def scale_temperature_dew(b, overwrite=True):
        """
        Scaling method for dew temperature
        """
        sf_P = iscale.get_scaling_factor(b.pressure, default=1e-5, warning=True)
        sf_mf = iscale.get_scaling_factor(b.mole_frac_comp, default=1e3, warning=True)

        for pp in b.params._pe_pairs:
            for j in b.component_list:
                (
                    l_phase,
                    v_phase,
                    vl_comps,
                    _,
                    _,
                    v_only_comps,
                ) = identify_VL_component_list(b, pp)
                if l_phase is None or v_phase is None:
                    continue
                elif v_only_comps != []:
                    continue

                if j in vl_comps:
                    sf = sf_P * sf_mf
                else:
                    sf = sf_mf

                # b.eq_temperature_dew is well-scaled by default
                iscale.constraint_scaling_transform(
                    b.eq_mole_frac_tdew[pp[0], pp[1], j], sf, overwrite=overwrite
                )

    # -------------------------------------------------------------------------
    # Bubble pressure methods
    @staticmethod
    def pressure_bubble(b):
        """
        Method for calculating bubble pressure
        """
        _non_vle_phase_check(b)
        try:

            def rule_bubble_press(b, p1, p2):
                (
                    l_phase,
                    v_phase,
                    vl_comps,
                    henry_comps,
                    _,
                    v_only_comps,
                ) = identify_VL_component_list(b, (p1, p2))

                if l_phase is None or v_phase is None:
                    # Not a VLE pair
                    return Constraint.Skip
                elif v_only_comps != []:
                    # Non-condensables present, no bubble point
                    return Constraint.Skip

                return b.pressure_bubble[p1, p2] == (
                    sum(b.mole_frac_comp[j] * b.pressure_sat_comp[j] for j in vl_comps)
                    + sum(
                        b.mole_frac_comp[j] * b.henry[l_phase, j] for j in henry_comps
                    )
                )

            b.eq_pressure_bubble = Constraint(
                b.params._pe_pairs, rule=rule_bubble_press
            )
        except AttributeError:
            b.del_component(b.eq_pressure_bubble)
            raise

        # Don't need a try/except here, will pass if first constraint did
        def rule_mole_frac_bubble_press(b, p1, p2, j):
            (
                l_phase,
                v_phase,
                vl_comps,
                henry_comps,
                _,
                v_only_comps,
            ) = identify_VL_component_list(b, (p1, p2))

            if l_phase is None or v_phase is None:
                # Not a VLE pair
                return Constraint.Skip
            elif v_only_comps != []:
                # Non-condensables present, no bubble point
                return Constraint.Skip

            if j in vl_comps:
                return (
                    b._mole_frac_pbub[p1, p2, j] * b.pressure_bubble[p1, p2]
                    == b.mole_frac_comp[j] * b.pressure_sat_comp[j]
                )
            if j in henry_comps:
                return (
                    b._mole_frac_pbub[p1, p2, j] * b.pressure_bubble[p1, p2]
                    == b.mole_frac_comp[j] * b.henry[l_phase, j]
                )
            else:
                return b._mole_frac_pbub[p1, p2, j] == 0

        b.eq_mole_frac_pbub = Constraint(
            b.params._pe_pairs, b.component_list, rule=rule_mole_frac_bubble_press
        )

    @staticmethod
    def scale_pressure_bubble(b, overwrite=True):
        """
        Scaling method for bubble pressure
        """
        sf_P = iscale.get_scaling_factor(b.pressure, default=1e-5, warning=True)
        sf_mf = iscale.get_scaling_factor(b.mole_frac_comp, default=1e3, warning=True)

        for pp in b.params._pe_pairs:
            for j in b.component_list:
                (
                    l_phase,
                    v_phase,
                    vl_comps,
                    _,
                    _,
                    v_only_comps,
                ) = identify_VL_component_list(b, pp)
                if l_phase is None or v_phase is None:
                    continue
                elif v_only_comps != []:
                    continue

                if j in vl_comps:
                    sf = sf_P * sf_mf
                else:
                    sf = sf_mf

                iscale.constraint_scaling_transform(
                    b.eq_pressure_bubble[pp[0], pp[1]], sf_P, overwrite=overwrite
                )
                iscale.constraint_scaling_transform(
                    b.eq_mole_frac_pbub[pp[0], pp[1], j], sf, overwrite=overwrite
                )

    # -------------------------------------------------------------------------
    # Dew pressure methods
    @staticmethod
    def pressure_dew(b):
        """
        Method for calculating dew pressure
        """
        _non_vle_phase_check(b)
        try:

            def rule_dew_press(b, p1, p2):
                (
                    l_phase,
                    v_phase,
                    vl_comps,
                    henry_comps,
                    l_only_comps,
                    _,
                ) = identify_VL_component_list(b, (p1, p2))

                if l_phase is None or v_phase is None:
                    # Not a VLE pair
                    return Constraint.Skip
                elif l_only_comps != []:
                    # Non-vaporisables present, no dew point
                    return Constraint.Skip

                return 0 == 1 - b.pressure_dew[p1, p2] * (
                    sum(b.mole_frac_comp[j] / b.pressure_sat_comp[j] for j in vl_comps)
                    + sum(
                        b.mole_frac_comp[j] / b.henry[l_phase, j] for j in henry_comps
                    )
                )

            b.eq_pressure_dew = Constraint(b.params._pe_pairs, rule=rule_dew_press)
        except AttributeError:
            b.del_component(b.eq_pressure_dew)
            raise

        # Don't need a try/except here, will pass if first constraint did
        def rule_mole_frac_dew_press(b, p1, p2, j):
            (
                l_phase,
                v_phase,
                vl_comps,
                henry_comps,
                l_only_comps,
                _,
            ) = identify_VL_component_list(b, (p1, p2))

            if l_phase is None or v_phase is None:
                # Not a VLE pair
                return Constraint.Skip
            elif l_only_comps != []:
                # Non-vaporisables present, no dew point
                return Constraint.Skip

            if j in vl_comps:
                return (
                    b._mole_frac_pdew[p1, p2, j] * b.pressure_sat_comp[j]
                    == b.mole_frac_comp[j] * b.pressure_dew[p1, p2]
                )
            elif j in henry_comps:
                return (
                    b._mole_frac_pdew[p1, p2, j] * b.henry[l_phase, j]
                    == b.mole_frac_comp[j] * b.pressure_dew[p1, p2]
                )
            else:
                return b._mole_frac_pdew[p1, p2, j] == 0

        b.eq_mole_frac_pdew = Constraint(
            b.params._pe_pairs, b.component_list, rule=rule_mole_frac_dew_press
        )

    @staticmethod
    def scale_pressure_dew(b, overwrite=True):
        """
        Scaling method for dew pressure
        """
        sf_P = iscale.get_scaling_factor(b.pressure, default=1e-5, warning=True)
        sf_mf = iscale.get_scaling_factor(b.mole_frac_comp, default=1e3, warning=True)

        for pp in b.params._pe_pairs:
            for j in b.component_list:
                (
                    l_phase,
                    v_phase,
                    vl_comps,
                    _,
                    _,
                    v_only_comps,
                ) = identify_VL_component_list(b, pp)
                if l_phase is None or v_phase is None:
                    continue
                elif v_only_comps != []:
                    continue

                if j in vl_comps:
                    sf = sf_P * sf_mf
                else:
                    sf = sf_mf

                # b.eq_pressure_dew is well-scaled by default
                iscale.constraint_scaling_transform(
                    b.eq_mole_frac_pdew[pp[0], pp[1], j], sf, overwrite=overwrite
                )


class LogBubbleDew:
    """General bubble and dew point calculations (log formulation)."""

    # -------------------------------------------------------------------------
    # Bubble temperature methods
    @staticmethod
    def temperature_bubble(b):
        """
        Method for constructing bubble temperature constraint
        """
        try:

            def rule_bubble_temp(b, p1, p2, j):
                (
                    l_phase,
                    v_phase,
                    vl_comps,
                    henry_comps,
                    _,
                    v_only_comps,
                ) = identify_VL_component_list(b, (p1, p2))

                if l_phase is None or v_phase is None:
                    # Not a VLE pair
                    return Constraint.Skip
                elif v_only_comps != []:
                    # Non-condensables present, no bubble point
                    return Constraint.Skip

                l_eos = b.params.get_phase(l_phase).config.equation_of_state
                v_eos = b.params.get_phase(v_phase).config.equation_of_state

                if j in vl_comps or j in henry_comps:
                    return l_eos.log_fug_phase_comp_Tbub(
                        b, l_phase, j, (p1, p2)
                    ) == v_eos.log_fug_phase_comp_Tbub(b, v_phase, j, (p1, p2))
                else:
                    return b._mole_frac_tbub[p1, p2, j] == 0

            b.eq_temperature_bubble = Constraint(
                b.params._pe_pairs, b.component_list, rule=rule_bubble_temp
            )
        except AttributeError:
            b.del_component(b.eq_temperature_bubble)
            raise

        # Don't need a try/except here, will pass if first constraint did
        def rule_mole_frac_bubble_temp(b, p1, p2):
            (
                l_phase,
                v_phase,
                vl_comps,
                henry_comps,
                _,
                v_only_comps,
            ) = identify_VL_component_list(b, (p1, p2))

            if l_phase is None or v_phase is None:
                # Not a VLE pair
                return Constraint.Skip
            elif v_only_comps != []:
                # Non-condensables present, no bubble point
                return Constraint.Skip

            return 1 == (
                sum(b._mole_frac_tbub[p1, p2, j] for j in vl_comps)
                + sum(b._mole_frac_tbub[p1, p2, j] for j in henry_comps)
            )

        b.eq_mole_frac_tbub = Constraint(
            b.params._pe_pairs, rule=rule_mole_frac_bubble_temp
        )

    @staticmethod
    def scale_temperature_bubble(b, overwrite=True):
        """
        Method for scaling bubble temperature
        """
        sf_mf = iscale.get_scaling_factor(b.mole_frac_comp, default=1e3, warning=True)

        for pp in b.params._pe_pairs:
            (
                l_phase,
                v_phase,
                _,
                _,
                _,
                v_only_comps,
            ) = identify_VL_component_list(b, pp)
            if l_phase is None or v_phase is None:
                continue
            elif v_only_comps != []:
                continue

            # Assume b.eq_temperature_bubble is well-scaled
            iscale.constraint_scaling_transform(
                b.eq_mole_frac_tbub[pp[0], pp[1]], sf_mf, overwrite=overwrite
            )

    # -------------------------------------------------------------------------
    # Dew temperature methods
    @staticmethod
    def temperature_dew(b):
        """
        Method for constructing dew temperature constraint
        """
        try:

            def rule_dew_temp(b, p1, p2, j):
                (
                    l_phase,
                    v_phase,
                    vl_comps,
                    henry_comps,
                    l_only_comps,
                    _,
                ) = identify_VL_component_list(b, (p1, p2))

                if l_phase is None or v_phase is None:
                    # Not a VLE pair
                    return Constraint.Skip
                elif l_only_comps != []:
                    # Non-vapouriszbles present, no dew point
                    return Constraint.Skip

                l_eos = b.params.get_phase(l_phase).config.equation_of_state
                v_eos = b.params.get_phase(v_phase).config.equation_of_state

                if j in vl_comps or j in henry_comps:
                    return l_eos.log_fug_phase_comp_Tdew(
                        b, l_phase, j, (p1, p2)
                    ) == v_eos.log_fug_phase_comp_Tdew(b, v_phase, j, (p1, p2))
                else:
                    return b._mole_frac_tdew[p1, p2, j] == 0

            b.eq_temperature_dew = Constraint(
                b.params._pe_pairs, b.component_list, rule=rule_dew_temp
            )
        except AttributeError:
            b.del_component(b.eq_temperature_dew)
            raise

        # Don't need a try/except here, will pass if first constraint did
        def rule_mole_frac_dew_temp(b, p1, p2):
            (
                l_phase,
                v_phase,
                vl_comps,
                henry_comps,
                l_only_comps,
                _,
            ) = identify_VL_component_list(b, (p1, p2))

            if l_phase is None or v_phase is None:
                # Not a VLE pair
                return Constraint.Skip
            elif l_only_comps != []:
                # Non-vaporisables present, no dew point
                return Constraint.Skip

            return 1 == (
                sum(b._mole_frac_tdew[p1, p2, j] for j in vl_comps)
                + sum(b._mole_frac_tdew[p1, p2, j] for j in henry_comps)
            )

        b.eq_mole_frac_tdew = Constraint(
            b.params._pe_pairs, rule=rule_mole_frac_dew_temp
        )

    @staticmethod
    def scale_temperature_dew(b, overwrite=True):
        """
        Method for scaling dew temperature
        """
        sf_mf = iscale.get_scaling_factor(b.mole_frac_comp, default=1e3, warning=True)

        for pp in b.params._pe_pairs:
            (
                l_phase,
                v_phase,
                _,
                _,
                _,
                v_only_comps,
            ) = identify_VL_component_list(b, pp)
            if l_phase is None or v_phase is None:
                continue
            elif v_only_comps != []:
                continue

            # Assume b.eq_temperature_dew is well-scaled
            iscale.constraint_scaling_transform(
                b.eq_mole_frac_tdew[pp[0], pp[1]], sf_mf, overwrite=overwrite
            )

    # -------------------------------------------------------------------------
    # Bubble pressure methods
    @staticmethod
    def pressure_bubble(b):
        """
        Method for constructing bubble pressure
        """
        try:

            def rule_bubble_press(b, p1, p2, j):
                (
                    l_phase,
                    v_phase,
                    vl_comps,
                    henry_comps,
                    _,
                    v_only_comps,
                ) = identify_VL_component_list(b, (p1, p2))

                if l_phase is None or v_phase is None:
                    # Not a VLE pair
                    return Constraint.Skip
                elif v_only_comps != []:
                    # Non-condensables present, no bubble point
                    return Constraint.Skip

                l_eos = b.params.get_phase(l_phase).config.equation_of_state
                v_eos = b.params.get_phase(v_phase).config.equation_of_state

                if j in vl_comps or j in henry_comps:
                    return l_eos.log_fug_phase_comp_Pbub(
                        b, l_phase, j, (p1, p2)
                    ) == v_eos.log_fug_phase_comp_Pbub(b, v_phase, j, (p1, p2))
                else:
                    return b._mole_frac_pbub[p1, p2, j] == 0

            b.eq_pressure_bubble = Constraint(
                b.params._pe_pairs, b.component_list, rule=rule_bubble_press
            )
        except AttributeError:
            b.del_component(b.eq_pressure_bubble)
            raise

        # Don't need a try/except here, will pass if first constraint did
        def rule_mole_frac_bubble_press(b, p1, p2):
            (
                l_phase,
                v_phase,
                vl_comps,
                henry_comps,
                _,
                v_only_comps,
            ) = identify_VL_component_list(b, (p1, p2))

            if l_phase is None or v_phase is None:
                # Not a VLE pair
                return Constraint.Skip
            elif v_only_comps != []:
                # Non-condensables present, no bubble point
                return Constraint.Skip

            return 1 == (
                sum(b._mole_frac_pbub[p1, p2, j] for j in vl_comps)
                + sum(b._mole_frac_pbub[p1, p2, j] for j in henry_comps)
            )

        b.eq_mole_frac_pbub = Constraint(
            b.params._pe_pairs, rule=rule_mole_frac_bubble_press
        )

    @staticmethod
    def scale_pressure_bubble(b, overwrite=True):
        """
        Method for scaling bubble pressure
        """
        sf_mf = iscale.get_scaling_factor(b.mole_frac_comp, default=1e3, warning=True)

        for pp in b.params._pe_pairs:
            (
                l_phase,
                v_phase,
                _,
                _,
                _,
                v_only_comps,
            ) = identify_VL_component_list(b, pp)
            if l_phase is None or v_phase is None:
                continue
            elif v_only_comps != []:
                continue

            # Assume b.eq_pressure_bubble is well-scaled
            iscale.constraint_scaling_transform(
                b.eq_mole_frac_pbub[pp[0], pp[1]], sf_mf, overwrite=overwrite
            )

    # -------------------------------------------------------------------------
    # Dew pressure methods
    @staticmethod
    def pressure_dew(b):
        """
        Method constructing dew pressure constraints
        """
        try:

            def rule_dew_press(b, p1, p2, j):
                (
                    l_phase,
                    v_phase,
                    vl_comps,
                    henry_comps,
                    _,
                    v_only_comps,
                ) = identify_VL_component_list(b, (p1, p2))

                if l_phase is None or v_phase is None:
                    # Not a VLE pair
                    return Constraint.Skip
                elif v_only_comps != []:
                    # Non-condensables present, no bubble point
                    return Constraint.Skip

                l_eos = b.params.get_phase(l_phase).config.equation_of_state
                v_eos = b.params.get_phase(v_phase).config.equation_of_state

                if j in vl_comps or j in henry_comps:
                    return l_eos.log_fug_phase_comp_Pdew(
                        b, l_phase, j, (p1, p2)
                    ) == v_eos.log_fug_phase_comp_Pdew(b, v_phase, j, (p1, p2))
                else:
                    return b._mole_frac_pdew[p1, p2, j] == 0

            b.eq_pressure_dew = Constraint(
                b.params._pe_pairs, b.component_list, rule=rule_dew_press
            )
        except AttributeError:
            b.del_component(b.eq_pressure_dew)
            raise

        # Don't need a try/except here, will pass if first constraint did
        def rule_mole_frac_dew_press(b, p1, p2):
            (
                l_phase,
                v_phase,
                vl_comps,
                henry_comps,
                l_only_comps,
                _,
            ) = identify_VL_component_list(b, (p1, p2))

            if l_phase is None or v_phase is None:
                # Not a VLE pair
                return Constraint.Skip
            elif l_only_comps != []:
                # Non-vaporisables present, no dew point
                return Constraint.Skip

            return 1 == (
                sum(b._mole_frac_pdew[p1, p2, j] for j in vl_comps)
                + sum(b._mole_frac_pdew[p1, p2, j] for j in henry_comps)
            )

        b.eq_mole_frac_pdew = Constraint(
            b.params._pe_pairs, rule=rule_mole_frac_dew_press
        )

    @staticmethod
    def scale_pressure_dew(b, overwrite=True):
        """
        Method for scaling dew pressure
        """
        sf_mf = iscale.get_scaling_factor(b.mole_frac_comp, default=1e3, warning=True)

        for pp in b.params._pe_pairs:
            (
                l_phase,
                v_phase,
                _,
                _,
                _,
                v_only_comps,
            ) = identify_VL_component_list(b, pp)
            if l_phase is None or v_phase is None:
                continue
            elif v_only_comps != []:
                continue

            # Assume b.eq_pressure_dew is well-scaled
            iscale.constraint_scaling_transform(
                b.eq_mole_frac_pdew[pp[0], pp[1]], sf_mf, overwrite=overwrite
            )


def _non_vle_phase_check(blk):
    if len(blk.phase_list) > 2:
        raise ConfigurationError(
            "Ideal assumption for calculating bubble and/or dew points is only valid "
            "for systems with two phases. Please use LogBubbleDew approach instead."
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2023 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Base class for custom scaling routines.

Author: Andrew Lee
"""
from copy import copy

from pyomo.environ import ComponentMap, units, value
from pyomo.core.base.units_container import UnitsError
from pyomo.core.expr import identify_variables
from pyomo.core.expr.calculus.derivatives import Modes, differentiate

from idaes.core.scaling.scaling_base import CONFIG, ScalerBase
from idaes.core.scaling.util import get_scaling_factor, NominalValueExtractionVisitor
import idaes.logger as idaeslog
from idaes.core.util.misc import StrEnum

# Set up logger
_log = idaeslog.getLogger(__name__)

CSCONFIG = CONFIG()

DEFAULT_UNIT_SCALING = {
    # "QuantityName: (reference units, scaling factor)
    # Model developers should be careful when using these, especially when
    # dealing with differential measurements (e.g. pressure and temperature differences)
    "Temperature": (units.K, 1e-2),
    "Pressure": (units.Pa, 1e-5),
}


class ConstraintScalingScheme(StrEnum):
    """
    Schemes available for calculating constraint scaling factors.

    * harmonicMean ('harmonic_mean'): sf = sum(1/abs(nominal value))
    * inverseSum ('inverse_sum'): sf = 1 / sum(abs(nominal value))
    * inverseRSS ('inverse_root_sum_squared'): sf = 1 / sqrt(sum(abs(nominal value)**2))
    * inverseMaximum ('inverse_maximum'): sf =  1 / max(abs(nominal value)
    * inverseMinimum ('inverse_minimum'): sf = 1 / min(abs(nominal value)
    """

    harmonicMean = "harmonic_mean"
    inverseSum = "inverse_sum"
    inverseRSS = "inverse_root_sum_squared"
    inverseMaximum = "inverse_maximum"
    inverseMinimum = "inverse_minimum"


class CustomScalerBase(ScalerBase):
    """
    Base class for custom scaling routines.

    """

    CONFIG = ScalerBase.CONFIG()

    # Common data structures for default scaling
    # DEFAULT_SCALING_FACTORS = {"component_local_name": DEFAULT_SCALING}
    DEFAULT_SCALING_FACTORS = None

    # UNIT_SCALING_FACTORS = {"units": UNIT_BASED_SCALING}
    UNIT_SCALING_FACTORS = copy(DEFAULT_UNIT_SCALING)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if self.DEFAULT_SCALING_FACTORS is not None:
            self.default_scaling_factors = copy(self.DEFAULT_SCALING_FACTORS)
        else:
            self.default_scaling_factors = {}

        if self.UNIT_SCALING_FACTORS is not None:
            self.unit_scaling_factors = copy(self.UNIT_SCALING_FACTORS)
        else:
            self.unit_scaling_factors = {}

    def scale_model(
        self,
        model,
        first_stage_fill_in: list = None,
        second_stage_fill_in: list = None,
        submodel_scalers: ComponentMap = None,
    ):
        """
        Default model scaling routine.

        This method performs a four-step scaling routine:

        1. Scale variables using variable_scaling_routine
        2. Perform first-stage scaling factor fill in using user provided method(s), called in order declared
        3. Scale constraints using constraint_scaling_routine
        4. Perform second-stage scaling factor fill in using user provided method(s), called in order declared

        Args:
            model: model to be scaled
            first_stage_fill_in: list of methods to use for first-stage scaling factor fill in
            second_stage_fill_in: list of methods to use for second-stage scaling factor fill in
            submodel_scalers: ComponentMap of Scalers to use for sub-models

        Returns:
            None
        """
        # Step 1: Call variable scaling routine
        self.variable_scaling_routine(
            model, overwrite=self.config.overwrite, submodel_scalers=submodel_scalers
        )
        # Step 2: Call variable fill in
        if first_stage_fill_in is not None:
            for i in first_stage_fill_in:
                i(model)

        # Step 3: Call constraint scaling routine
        self.constraint_scaling_routine(
            model, overwrite=self.config.overwrite, submodel_scalers=submodel_scalers
        )

        # Step 4: Call constraint fill in
        if second_stage_fill_in is not None:
            for i in second_stage_fill_in:
                i(model)

    def variable_scaling_routine(
        self, model, overwrite: bool = False, submodel_scalers: ComponentMap = None
    ):
        """
        Routine to apply scaling factors to variables in model.

        Derived classes must overload this method.

        Args:
            model: model to be scaled
            overwrite: whether to overwrite existing scaling factors
            submodel_scalers: ComponentMap of Scalers to use for sub-models

        Returns:
            None
        """
        raise NotImplementedError(
            "Custom Scaler has not implemented a variable_scaling_routine method."
        )

    def constraint_scaling_routine(
        self, model, overwrite: bool = False, submodel_scalers: ComponentMap = None
    ):
        """
        Routine to apply scaling factors to constraints in model.

        Derived classes must overload this method.

        Args:
            model: model to be scaled
            overwrite: whether to overwrite existing scaling factors
            submodel_scalers: ComponentMap of Scalers to use for sub-models

        Returns:
            None
        """
        raise NotImplementedError(
            "Custom Scaler has not implemented a constraint_scaling_routine method."
        )

    def get_default_scaling_factor(self, component):
        """
        Get scaling factor for component from dict of default values.

        Args:
            component: component to get default scaling factor for

        Returns:
            default scaling factor if it exists, else None
        """
        try:
            return self.default_scaling_factors[component.local_name]
        except KeyError:
            # Might be indexed, see if parent component has default scaling
            parent = component.parent_component()
            try:
                return self.default_scaling_factors[parent.local_name]
            except KeyError:
                # No default scaling found, give up
                pass

            # Log a message and return nothing
            _log.debug(f"No default scaling factor found for {component.name}")
            return None

    # Common methods for variable scaling
    def scale_variable_by_component(
        self, target_variable, scaling_component, overwrite: bool = False
    ):
        """
        Set scaling factor for target_variable equal to that of scaling_component.

        Args:
            target_variable: variable to set scaling factor for
            scaling_component: component to use for scaling factor
            overwrite: whether to overwrite existing scaling factors

        Returns:
            None
        """
        sf = get_scaling_factor(scaling_component)

        if sf is not None:
            self.set_variable_scaling_factor(
                variable=target_variable, scaling_factor=sf, overwrite=overwrite
            )
        else:
            _log.debug(
                f"Could not set scaling factor for {target_variable.name}, "
                f"no scaling factor set for {scaling_component.name}"
            )

    def scale_variable_by_bounds(self, variable, overwrite: bool = False):
        """
        Set scaling factor for variable based on bounds.

        If variable has both upper and lower bounds, scaling factor will be based on the
        mean of the bounds. If variable has only one bound, scaling factor will be based
        on that bound. If variable has no bounds, scaling factor will not be set.

        Args:
            variable: variable to set scaling factor for
            overwrite: whether to overwrite existing scaling factors

        Returns:
            None
        """
        if variable.lb is not None:
            if variable.ub is not None:
                # Both bounds, use mean
                xmag = 0.5 * (variable.ub + variable.lb)
            else:
                # Only lower bound
                xmag = variable.lb
        elif variable.ub is not None:
            # Only upper bound
            xmag = variable.ub
        else:
            # No bounds
            _log.debug(
                f"No scaling factor set for {variable.name}; variable has no bounds."
            )
            return

        if xmag == 0:
            sf = 1
        else:
            sf = 1 / abs(xmag)

        self.set_variable_scaling_factor(
            variable=variable, scaling_factor=sf, overwrite=overwrite
        )

    def scale_variable_by_default(self, variable, overwrite: bool = False):
        """
        Set scaling factor for variable based on default scaling factor.

        Args:
            variable: variable to set scaling factor for
            overwrite: whether to overwrite existing scaling factors

        Returns:
            None
        """
        sf = self.get_default_scaling_factor(variable)
        if sf is not None:
            self.set_variable_scaling_factor(
                variable=variable, scaling_factor=sf, overwrite=overwrite
            )
        else:
            _log.debug(
                f"Could not set scaling factor for {variable.name}, "
                f"no default scaling factor set."
            )

    def scale_variable_by_units(self, variable, overwrite: bool = False):
        """
        Set scaling factor for variable based on units of measurement.

        Units of measurement for variable are compared to those stored in
        self.unit_scaling_factors, and if a match is found the scaling factor set
        using the associated value.

        Args:
            variable: variable to set scaling factor for
            overwrite: whether to overwrite existing scaling factors

        Returns:
            None
        """
        uom = units.get_units(variable)

        sf = None
        # Keys in self.unit_scaling_factors are not used - only required because Pyomo
        # units are non-hashable and thus cannot be keys
        for refunits, unit_scale in self.unit_scaling_factors.values():
            try:
                # Try convert the reference scaling factor to variable units
                # TODO: Have not found a more efficient way to do this, as Pyomo basically
                # TODO: involves a try/except anyway
                # Need to invert reference scaling factor, and then invert result
                sf = 1 / units.convert_value(
                    1 / unit_scale, from_units=refunits, to_units=uom
                )
                # Break once we have a match - no need to continue
                break
            except UnitsError:
                pass

        if sf is not None:
            self.set_variable_scaling_factor(
                variable=variable,
                scaling_factor=sf,
                overwrite=overwrite,
            )
        else:
            _log.debug(
                f"No scaling factor set for {variable.name}; no match for units {uom} found "
                "in self.unit_scaling_factors"
            )

    # Common methods for constraint scaling
    def scale_constraint_by_component(
        self, target_constraint, scaling_component, overwrite: bool = False
    ):
        """
        Set scaling factor for target_constraint equal to that of scaling_component.

        Args:
            target_constraint: constraint to set scaling factor for
            scaling_component: component to use for scaling factor
            overwrite: whether to overwrite existing scaling factors

        Returns:
            None
        """
        sf = get_scaling_factor(scaling_component)
        if sf is not None:
            self.set_constraint_scaling_factor(
                constraint=target_constraint, scaling_factor=sf, overwrite=overwrite
            )
        else:
            _log.debug(
                f"Could not set scaling factor for {target_constraint.name}, "
                f"no scaling factor set for {scaling_component.name}"
            )

    def scale_constraint_by_default(self, constraint, overwrite: bool = False):
        """
        Set scaling factor for constraint based on default scaling factor.

        Args:
            constraint: constraint to set scaling factor for
            overwrite: whether to overwrite existing scaling factors

        Returns:
            None
        """
        sf = self.get_default_scaling_factor(constraint)
        if sf is not None:
            self.set_constraint_scaling_factor(
                constraint=constraint, scaling_factor=sf, overwrite=overwrite
            )
        else:
            _log.debug(
                f"Could not set scaling factor for {constraint.name}, "
                f"no default scaling factor set."
            )

    def get_expression_nominal_values(self, expression):
        """
        Calculate nominal values for each additive term in a Pyomo expression.

        The nominal value of any Var is defined as the inverse of its scaling factor
        (if assigned, else 1).

        Args:
            expression: Pyomo expression to collect nominal values for

        Returns:
            list of nominal values for each additive term
        """
        # For convenience, if expression is a Pyomo component with an expr attribute,
        # redirect to the expr attribute
        if hasattr(expression, "expr"):
            expression = expression.expr

        return NominalValueExtractionVisitor().walk_expression(expression)

    def scale_constraint_by_nominal_value(
        self,
        constraint,
        scheme: ConstraintScalingScheme = ConstraintScalingScheme.inverseMaximum,
        overwrite: bool = False,
    ):
        """
        Set scaling factor for constraint based on the nominal value(s).

        Terms with expected magnitudes of 0 will be ignored.

        Args:
            constraint: constraint to set scaling factor for
            scheme: ConstraintScalingScheme Enum indicating method to apply
              for determining constraint scaling.
            overwrite: whether to overwrite existing scaling factors

        Returns:
            None
        """
        nominal = self.get_expression_nominal_values(constraint.expr)

        # Remove any 0 terms
        nominal = [j for j in nominal if j != 0]

        if len(nominal) == 0:
            # No non-zero terms...
            sf = 1
        elif scheme == ConstraintScalingScheme.harmonicMean:
            sf = sum(1 / abs(i) for i in nominal)
        elif scheme == ConstraintScalingScheme.inverseSum:
            sf = 1 / sum(abs(i) for i in nominal)
        elif scheme == ConstraintScalingScheme.inverseRSS:
            sf = 1 / sum(abs(i) ** 2 for i in nominal) ** 0.5
        elif scheme == ConstraintScalingScheme.inverseMaximum:
            sf = 1 / max(abs(i) for i in nominal)
        elif scheme == ConstraintScalingScheme.inverseMinimum:
            sf = 1 / min(abs(i) for i in nominal)
        else:
            raise ValueError(
                f"Invalid value for 'scheme' argument ({scheme}) in "
                "scale_constraint_by_nominal_value."
            )

        self.set_constraint_scaling_factor(
            constraint=constraint, scaling_factor=sf, overwrite=overwrite
        )

    def scale_constraint_by_nominal_derivative_norm(
        self, constraint, norm: int = 2, overwrite: bool = False
    ):
        """
        Scale constraint by norm of partial derivatives.

        Calculates partial derivatives of constraint at nominal variable values,
        and then scaled the constraint by the user-selected norm of these derivatives.
        Given perfect variable scaling, this should provide a similar result to
        applying scaling based on the Jacobian norm, however this approach does not
        require an initial solution for the problem (relying on nominal values instead).

        Args:
            constraint: constraint to be scaled.
            norm: type of norm to use for scaling. Must be a positive integer.
            overwrite: whether to overwrite existing scaling factors.

        Returns:
            None
        """
        # Cast norm to int to make sure it is valid
        norm = int(norm)
        if norm < 1:
            raise ValueError(f"norm must be a positive integer (received {norm})")

        var_data = []
        try:
            # Iterate over all variables in constraint
            for v in identify_variables(constraint.body):
                # Store current value for restoration
                ov = v.value  # original value
                sf = self.get_scaling_factor(v)  # scaling factor
                if sf is None:
                    # If no scaling factor set, use nominal value of 1
                    sf = 1

                var_data.append((v, ov, sf))

            # Get partial derivatives
            pjac = []
            for v in var_data:
                # Iterate over all variable and set values
                for w in var_data:
                    if w is not v:
                        # Set all other variables to their nominal magnitude
                        # nominal_value = 1/ scaling_factor
                        w[0].value = 1 / w[2]
                    else:
                        # Set derivative var to scaled value of 1
                        # With perfect scaling, scaling_factor * value = 1
                        w[0].value = 1

                pjac.append(
                    value(
                        differentiate(
                            expr=constraint.body, wrt=v[0], mode=Modes.reverse_symbolic
                        )
                        * (1 / v[2])  # Need to divide by scaling_factor
                    )
                )

        finally:
            # Restore all values for clean up
            for v in var_data:
                v[0].value = v[1]

        # Calculate norm
        cnorm = sum(abs(j) ** norm for j in pjac) ** (1 / norm)
        if cnorm != 0:
            sf = 1 / cnorm
        else:
            sf = 1
        self.set_constraint_scaling_factor(constraint, sf, overwrite=overwrite)

    # Other methods
    def propagate_state_scaling(
        self, target_state, source_state, overwrite: bool = False
    ):
        """
        Propagate scaling of state variables from one StateBlock to another.

        Indexing of target and source StateBlocks must match.

        Args:
            target_state: StateBlock to set scaling factors on
            source_state: StateBlock to use as source for scaling factors
            overwrite: whether to overwrite existing scaling factors

        Returns:
            None
        """
        for bidx, target_data in target_state.items():
            target_vars = target_data.define_state_vars()
            source_vars = source_state[bidx].define_state_vars()

            for state, var in target_vars.items():
                for vidx, vardata in var.items():
                    self.scale_variable_by_component(
                        target_variable=vardata,
                        scaling_component=source_vars[state][vidx],
                        overwrite=overwrite,
                    )

    def call_submodel_scaler_method(
        self,
        submodel,
        method: str,
        submodel_scalers: ComponentMap = None,
        overwrite: bool = False,
    ):
        """
        Call scaling method for submodel.

        Scaler for submodel is taken from submodel_scalers if present, otherwise the
        default scaler for the submodel is used.

        Args:
            submodel: submodel to be scaled
            submodel_scalers: user provided ComponentMap of Scalers to use for submodels
            method: name of method to call from submodel (as string)
            overwrite: whether to overwrite existing scaling factors

        Returns:
            None
        """
        if submodel_scalers is None:
            submodel_scalers = {}

        # Iterate over indices of submodel
        for smdata in submodel.values():
            # Get Scaler for submodel
            if submodel in submodel_scalers:
                scaler = submodel_scalers[submodel]
                if callable(scaler):
                    # Check to see if Scaler is callable - this implies it is a class and not an instance
                    # Call the class to create an instance
                    scaler = scaler()
                _log.debug(f"Using user-defined Scaler for {submodel}.")
            else:
                try:
                    scaler = smdata.default_scaler
                    _log.debug(f"Using default Scaler for {submodel}.")
                except AttributeError:
                    _log.debug(
                        f"No default Scaler set for {submodel}. Cannot call {method}."
                    )
                    return
                if scaler is not None:
                    scaler = scaler()
                else:
                    _log.debug(f"No Scaler found for {submodel}. Cannot call {method}.")

            # If a Scaler is found, call desired method
            if scaler is not None:
                try:
                    smeth = getattr(scaler, method)
                except AttributeError:
                    raise AttributeError(
                        f"Scaler for {submodel} does not have a method named {method}."
                    )
                smeth(smdata, overwrite=overwrite)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import numpy as np
from copy import deepcopy

from ..util.util import myArrayEq, myPointsEq, ListHasPoint
from .parsers.PDB import readPointsAndAtomsFromPDB
from .parsers.XYZ import readPointsAndAtomsFromXYZ
from .parsers.CFG import readPointsAndAtomsFromCFG
from .geometry import RectPrism


class Canvas(object):
    """A class for combining geometric points and neighbors.

    This class contains a list of Cartesian points coupled with a graph of nodes for sites and arcs
    for bonds. A ``Canvas`` object establishes a mapping from the abstract, mathematical modeling of
    materials as graphs to the geometry of the material lattice. The list of points and neighbor
    connections necessary to create a ``Canvas`` object can be obtained from the combination of
    ``Lattice``, ``Shape``, and ``Tiling`` objects.
    """

    DBL_TOL = 1e-5

    # === STANDARD CONSTRUCTOR
    def __init__(self, Points=None, NeighborhoodIndexes=None, DefaultNN=0):
        if Points is None and NeighborhoodIndexes is None:
            Points = []
            NeighborhoodIndexes = []
        elif Points is None:
            Points = [None] * len(NeighborhoodIndexes)
        elif NeighborhoodIndexes is None:
            NeighborhoodIndexes = [[None] * DefaultNN for _ in range(len(Points))]
        self._Points = Points
        self._NeighborhoodIndexes = NeighborhoodIndexes
        self.__DefaultNN = DefaultNN
        assert self.isConsistentWithDesign()

    # === CONSTRUCTOR - From PDB File
    @classmethod
    def fromPDB(cls, filename, Lat=None, DefaultNN=0):
        """Make Canvas by reading from PDB file.

        Args:
            filename(str): Name of PDB file to read.
            Lat (Lattice, optional): A lattice to define neighbor connections
            DefaultNN(int, optional): Number of neighbors to allocate
                in neighborhood. (Default value = 0)

        Returns:
            Canvas: A new Canvas object.

        """
        Pts, _ = readPointsAndAtomsFromPDB(filename)
        result = cls(Points=Pts, DefaultNN=DefaultNN)
        if Lat is not None:
            result.setNeighborsFromFunc(Lat.getNeighbors)
        return result

    @classmethod
    def fromXYZ(cls, filename, Lat=None, DefaultNN=0):
        """Make Canvas by reading from XYZ file.

        Args:
            filename(str): Name of XYZ file to read.
            Lat (Lattice, optional): A lattice to define neighbor connections
            DefaultNN(int, optional): Number of neighbors to allocate
                in neighborhood. (Default value = 0)

        Returns:
            Canvas: A new Canvas object.

        """
        Pts, _ = readPointsAndAtomsFromXYZ(filename)
        result = cls(Points=Pts, DefaultNN=DefaultNN)
        if Lat is not None:
            result.setNeighborsFromFunc(Lat.getNeighbors)
        return result

    @classmethod
    def fromCFG(cls, filename, Lat=None, DefaultNN=0):
        """Make Canvas by reading from CFG file.

        Args:
            filename(str): Name of CFG file to read.
            Lat (Lattice, optional): A lattice to define neighbor connections
            DefaultNN(int, optional): Number of neighbors to allocate
                in neighborhood. (Default value = 0)

        Returns:
            Canvas: A new Canvas object.

        """
        Pts, _ = readPointsAndAtomsFromCFG(filename)
        result = cls(Points=Pts, DefaultNN=DefaultNN)
        if Lat is not None:
            result.setNeighborsFromFunc(Lat.getNeighbors)
        return result

    # === CONSTRUCTOR - From Lattice and Shape
    @classmethod
    def fromLatticeAndShape(
        cls, Lat, S, Seed=np.array([0, 0, 0], dtype=float), DefaultNN=0
    ):
        """Make Canvas by iterating over Lattice points that fit in Shape.

        This constructor starts from a seed location and repeatedly adds
        neighbors until there are no more neighbors that lie inside the
        provided Shape object. It is potentially very slow and should only
        be considered if other methods are unavailable.
        Prefer Canvas.fromLatticeAndShapeScan.

        Args:
            DefaultNN:
            Lat(Lattice): Lattice to provide getNeighbors function.
            S(Shape): Shape to iterate over.
            Seed(numpy.ndarray, optional): Location to begin adding neighbors from.
                Should be on the Lattice. (Default value =
                np.array([0,0,0],dtype=float))
            DefaultNN(int, optional): Number of neighbors to allocate
                in neighborhood. (Default value = 0)

        Returns:
            Canvas: A new Canvas object.

        """
        result = cls(DefaultNN=DefaultNN)
        Stack = [Seed]
        while len(Stack) > 0:
            P = Stack.pop()
            if not result.hasPoint(P) and P in S:
                PNs = Lat.getNeighbors(P)
                result.addLocation(P, len(PNs))
                # NOTE: At first, we checked if P needed to be added
                #       (i.e., if it was not already in the Stack)
                #       but doing so was significantly slower than
                #       just extending the Stack without checks and
                #       just checking if P was already in the result
                Stack.extend(PNs)
        result.setNeighborsFromFunc(Lat.getNeighbors)
        return result

    @classmethod
    def fromLatticeAndShapeScan(cls, Lat, argPolyhedron, DefaultNN=0):
        """Make Canvas by iterating over Lattice points that fit in Shape.

        This constructor takes advantage of methods in Lattice to
        efficiently scan over sites. This requires the Lattice.Scan
        method to produce a generator object.

        Args:
            DefaultNN:
            Lat(Lattice): Lattice with has defined Scan method.
            argPolyhedron(Polyhedron: Polyhedron): Shape to iterate over.
                NOTE: A Polyhedron is required because there is a
                complicated step in finding bounds for the Shape
                in the reference lattice space that is not
                generally valid for all Shapes.
            DefaultNN(int, optional): Number of neighbors to allocate
                in neighborhood. (Default value = 0)

        Returns:
            Canvas: A new Canvas object.

        """
        result = cls(DefaultNN=DefaultNN)
        BBox = RectPrism.fromPointsBBox(argPolyhedron.getBounds())
        for P in Lat.Scan(BBox):
            if P in argPolyhedron:
                result.addLocation(P, len(Lat.getNeighbors(P)))
        result.setNeighborsFromFunc(Lat.getNeighbors)
        return result

    @classmethod
    def fromLatticeAndTiling(
        cls, Lat, T, Seed=np.array([0, 0, 0], dtype=float), DefaultNN=0
    ):
        """Make Canvas by iterating over Lattice points that fit in Tiling.

        See documentation for fromLatticeAndShape.
        This constructor additionally makes the resulting Canvas periodic.

        Args:
            DefaultNN:
            Lat(Lattice): Lattice to provide getNeighbors function.
            T(Tiling): Tiling which provides a Shape to iterate over.
            Seed(numpy.ndarray, optional): Location to begin adding neighbors from.
                Should be on the Lattice. (Default value = np.array([0,0,0],dtype=float))
            DefaultNN(int, optional): Number of neighbors to allocate
                in neighborhood. (Default value = 0)

        Returns:
            Canvas: A new Canvas object.

        """
        result = cls.fromLatticeAndShape(
            Lat, T.TileShape, Seed=Seed, DefaultNN=DefaultNN
        )
        result.makePeriodic(T, Lat.getNeighbors)
        return result

    @classmethod
    def fromLatticeAndTilingScan(cls, Lat, T, DefaultNN=0):
        """Make Canvas by iterating over Lattice points that fit in Tiling.

        See documentation for fromLatticeAndTiling.
        This constructor additionally makes the resulting Canvas periodic.

        Args:
            Lat(Lattice): Lattice with has defined Scan method.
            T(Tiling): Tiling that provides a Polyhedron shape.
            DefaultNN(int, optional): Number of neighbors to allocate
                in neighborhood. (Default value = 0)

        Returns:
            Canvas: A new Canvas object.

        """
        result = cls.fromLatticeAndShapeScan(Lat, T.TileShape, DefaultNN=DefaultNN)
        result.makePeriodic(T, Lat.getNeighbors)
        return result

    # === ASSERTION OF CLASS DESIGN
    def isConsistentWithDesign(self):
        """Determine if object is consistent with class assumptions."""
        if len(self.Points) != len(self.NeighborhoodIndexes):
            return False
        return True

    # === MANIPULATION METHODS
    def addLocation(self, P, NNeighbors=None):
        """Add new location to Points.

        Args:
            P(numpy.ndarray): Point to add.
            NNeighbors(int, optional): Number of neighbors to allocate
                in a neighborhood. If None, use the instance default
                self.DefaultNN (Default value = None)

        Returns:
            None.

        """
        assert not self.hasPoint(P)
        self._Points.append(P)
        self._NeighborhoodIndexes.append([None] * (NNeighbors or self.__DefaultNN))
        assert self.isConsistentWithDesign()

    def setNeighbors(self, P1, P2, l=None):
        """Set a (directed) neighbor connection between two points.

        Args:
            P1(numpy.ndarray): Canvas Point to set neighbor for.
            P2(numpy.ndarray): Canvas Point to set as neighbor.
            l(int, optional): Index of neighbor to set. For example, if
                l=3, then P2 is set to the fourth neighbor of P1.
                If None, appends to the neighborhood.
                (Default value = None)

        Returns:
            None.

        """
        assert self.hasPoint(P1)
        assert self.hasPoint(P2)
        i = self.getPointIndex(P1)
        j = self.getPointIndex(P2)
        self.setNeighborsIJ(i, j, l=l)
        assert self.isConsistentWithDesign()

    def setNeighborsIJ(self, i, j, l=None):
        """Set a (directed) neighbor connection between two indices.

        Args:
            i(int): Index of location to set neighbor for.
            j(int): Index of location to set as neighbor.
            l(int, optional): Index of neighbor to set. For example, if
                l=3, then j is set to the fourth neighbor of i.
                If None, appends to the neighborhood.
                (Default value = None)

        Returns:
            None.

        """
        if l is None:
            self._NeighborhoodIndexes[i].append(j)
        else:
            self._NeighborhoodIndexes[i][l] = j
        assert self.isConsistentWithDesign()

    def setNeighborLofI(self, PN, l, i, blnSetNoneOtherwise=True):
        """Set a (directed) neighbor connection between a Point and index.

        Args:
            blnSetNoneOtherwise:
            i:
            PN(numpy.ndarray): Canvas Point to set as neighbor.
            l(int): Index of neighbor to set. For example, if
                l=3, then PN is set as the fourth neighbor of i.
            i(int): Index of location to set neighbor for.
            blnSetNoneOtherwise(bool, optional): Flag to control behavior
                if PN is not found in Canvas. (Default value = True)

        Returns:
            None.

        """
        assert i < len(self._NeighborhoodIndexes)
        assert l < len(self._NeighborhoodIndexes[i])
        if self.hasPoint(PN):
            self._NeighborhoodIndexes[i][l] = self.getPointIndex(PN)
        elif blnSetNoneOtherwise:
            self._NeighborhoodIndexes[i][l] = None
        assert self.isConsistentWithDesign()

    def setNeighborsOfI(self, PNs, i):
        """Set a list of points as neighbors to an index.

        Args:
            PNs(list<numpy.ndarray>): Points to set as neighbors of i.
            i(int): Index to set neighbors for.

        Returns:
            None.

        """
        self._NeighborhoodIndexes[i] = [None] * len(PNs)
        for l, P in enumerate(PNs):
            self.setNeighborLofI(P, l, i)
        assert self.isConsistentWithDesign()

    def setNeighborsFromFunc(self, NeighborsFunc):
        """Set neighbors across the Canvas from a functor.

        Args:
            NeighborsFunc(function): Function that takes as input a
                point (numpy.ndarray) and returns a list of points.

        Returns:
            None.

        """
        for i, P in enumerate(self.Points):
            PNs = NeighborsFunc(P)
            self.setNeighborsOfI(PNs, i)
        assert self.isConsistentWithDesign()

    def getNeighborsFromFunc(self, NeighborsFunc, layer=1):
        """Get neighbors across the Canvas from a functor.

        Args:
            NeighborsFunc(function): Function that takes as input a point
                (numpy.ndarray) and returns a list of points.
            layer(int): Target layer of neighbors.

        Returns:
            list<list<int>> Indexes matrix

        """
        result = [[] for _ in range(len(self.Points))]
        for i, P in enumerate(self.Points):
            PNs = NeighborsFunc(P, layer)
            result[i] = [None] * len(PNs)
            for j, PN in enumerate(PNs):
                if self.hasPoint(PN):
                    result[i][j] = self.getPointIndex(PN)
        return result

    def getNeighborsFromFuncAndTiling(self, NeighborsFunc, argTiling, layer=1):
        """Get neighbors across the Canvas from a functor and Tiling.

        Args:
            NeighborsFunc(function): Function that takes as input a point
                (numpy.ndarray) and returns a list of points.
            argTiling(Tiling): Tiling object that specifies the periodicity
                of the Canvas.
            layer(int): Target layer of neighbors.

        Returns:
            list<list<int>> Indexes matrix
        """
        result = [[] for _ in range(len(self.Points))]
        for i, P in enumerate(self.Points):
            PNs = NeighborsFunc(P, layer)
            result[i] = [None] * len(PNs)
            for j, PN in enumerate(PNs):
                if self.hasPoint(PN):
                    result[i][j] = self.getPointIndex(PN)
                else:
                    for TilingDirection in argTiling.TilingDirections:
                        PtoTry = PN + TilingDirection
                        if self.hasPoint(PtoTry):
                            result[i][j] = self.getPointIndex(PtoTry)
        return result

    def makePeriodic(self, argTiling, NeighborsFunc):
        """Make connections periodic across the edges of Tiling.

        Args:
            argTiling(Tiling: Tiling): Tiling to provide TilingDirections.
            NeighborsFunc(function): Function that takes as input a
                point (numpy.ndarray) and returns a list of points.

        Returns:
            None.

        """
        for i, P in enumerate(self.Points):
            LatNeighbors = NeighborsFunc(P)
            for l, Index in enumerate(self.NeighborhoodIndexes[i]):
                if Index is None:
                    assert not self.hasPoint(
                        LatNeighbors[l]
                    )  # else, Canvas constructed incorrectly
                    for TilingDirection in argTiling.TilingDirections:
                        PtoTry = LatNeighbors[l] + TilingDirection
                        if self.hasPoint(PtoTry):
                            self._NeighborhoodIndexes[i][l] = self.getPointIndex(PtoTry)
                            break

    def addShells(self, n, NeighborsFunc):
        """Add locations in n-shells around current Points.

        Args:
            n(int): Number of shells to add.
            NeighborsFunc(function): Function that takes as input a
                point (numpy.ndarray) and returns a list of points.

        Returns:
            None.

        """
        for _ in range(n):
            self.addShell(NeighborsFunc)

    def addShell(self, NeighborsFunc):
        """Add locations in a shell around current Points.

        Args:
            NeighborsFunc(function): Function that takes as input a
                point (numpy.ndarray) and returns a list of points.

        Returns:
            None.

        """
        Shell = self.getShell(NeighborsFunc)
        for P in Shell:
            self.addLocation(P)
        self.setNeighborsFromFunc(NeighborsFunc)

    def transform(self, TransF):
        """Transform the Points in Canvas according to a functor.

        Args:
            TransF(TransformFunc): Transformation to apply to Points.

        Returns:
            None.

        """
        for P in self._Points:
            TransF.transform(P)

    def getTransformed(self, TransF):
        """Copy and transform this Canvas.

        Args:
            TransF(TransformFunc): Transformation to apply to Points.

        Returns:
            None.

        """
        result = deepcopy(self)
        result.transform(TransF)
        return result

    def addOther(self, other, blnAssertNotAlreadyInCanvas=True):
        """Add other Canvas to this one.

        Args:
            other(Canvas): Canvas to append.
            blnAssertNotAlreadyInCanvas(bool, optional): Flag to enable
                assertion that all locations were new and unique.
                (Default value = True)

        Returns:
            None.

        """
        for P in other.Points:
            if blnAssertNotAlreadyInCanvas:
                assert not self.hasPoint(P)
            self.addLocation(P)

    # === PROPERTY EVALUATION METHODS
    def __len__(self):
        """Get the number of Points for this Canvas."""
        return len(self.Points)

    def __eq__(self, other):
        """Compare strict equality of Canvas data."""
        return (
            myPointsEq(self.Points, other.Points, Canvas.DBL_TOL)
            and self.NeighborhoodIndexes == other.NeighborhoodIndexes
        )

    def hasPoint(self, P):
        """Identify if point is in Canvas.

        Args:
            P(numpy.ndarray): Point to identify in Canvas.

        Returns:
            bool) True if Points has P.

        """
        for Q in self.Points:
            # NOTE: There are several ways to test point membership.
            #       This is optimized for speed.
            if myArrayEq(P, Q, Canvas.DBL_TOL):
                return True
        return False

    def getPointIndex(self, P):
        """Identify the index of a point in the Canvas.

        Args:
            P(numpy.ndarray): Point in the Canvas.

        Returns:
            int) Index of P in Points.

        """
        for i, Q in enumerate(self.Points):
            # NOTE: There are several ways to test point membership.
            #       This is optimized for speed.
            if myArrayEq(P, Q, Canvas.DBL_TOL):
                return i
        return None

    def getNeighbors(self, P):
        """Identify set of neighbors to a point in Canvas.

        Args:
            P(numpy.ndarray): Point to get neighbors for.

        Returns:
            list<int>: Neighborhood of P.

        """
        return self.NeighborhoodIndexes[self.getPointIndex(P)]

    def getNeighborLofI(self, l, i):
        """Identify neighbor for specific index and neighbor order.

        Args:
            l(int): Order of neighbor in neighborhood
            i(int): Index to consider the neighborhood of.

        Returns:
            int: Index for neighbor l of i.

        """
        assert i < len(self.NeighborhoodIndexes)
        assert l < len(self.NeighborhoodIndexes[i])
        return self.NeighborhoodIndexes[i][l]

    def getShell(self, NeighborsFunc):
        """Identify set of neighboring points not in Canvas.

        Args:
            NeighborsFunc(function): Function that takes as input a
                point (numpy.ndarray) and returns a list of points.

        Returns:
            list<numpy.ndarray>: Set of points to consider as
            neighboring shell.

        """
        result = []
        for i, P in enumerate(self.Points):
            Neighs = NeighborsFunc(P)
            for Neigh in Neighs:
                if not self.hasPoint(Neigh) and not ListHasPoint(
                    result, Neigh, Canvas.DBL_TOL
                ):
                    result.append(Neigh)
        return result

    def getNeighborhoodIndexes(self, Lat, layer=1, T=None):
        """Wrapper of functions that returns neighbors across the Canvas

        Args:
            Lat(Lattice): Lattice of the Canvas
            layer(int): optional, target layer of neighbors
            T(Tiling): optional, Tiling object that specifies the periodicity

        Returns:
            list<list<int>> Indexes matrix
        """
        if layer == 1:
            return self.NeighborhoodIndexes
        elif T is None:
            return self.getNeighborsFromFunc(Lat.getNeighbors, layer)
        else:
            return self.getNeighborsFromFuncAndTiling(Lat.getNeighbors, T, layer)

    # === BASIC QUERY METHODS
    @property
    def Points(self):
        """Get Points in Canvas."""
        return self._Points

    @property
    def NeighborhoodIndexes(self):
        """Get description of neighborhoods in Canvas."""
        return self._NeighborhoodIndexes

    # === REPORTING METHODS
    def printPoints(self):
        """Pretty-print Points."""
        for i, P in enumerate(self.Points):
            print("{}: {}".format(i, P))

    def printNeighborhoodIndexes(self, layer=1):
        """Pretty-print NeighborhoodIndexes."""
        for i, Neighborhood in enumerate(self.getNeighborhoodIndexes(layer)):
            print("{}: {}".format(i, Neighborhood))
